from goopylib.util import GraphicsError

from goopylib.objects.GraphicsObject import GraphicsObject
from goopylib.objects.Line import Line
from goopylib.Point import Point

class Slider:
    def __init__(self, p, length, orientation, minimum, maximum, value, pointer, step=1, style=None, outline=None,
                 width=None, text=None, label="Value "):

        self.range = [minimum - 1, maximum + 1]
        self.state = value - minimum

        if not self.range[1] > value > self.range[0]:
            raise GraphicsError(
                f"\n\nSlide Bar value must between min/max values: {minimum} < value ({value}) < {maximum}")

        self.anchor = p.clone()

        self.maximum = maximum
        self.minimum = minimum

        if self.minimum > self.maximum:
            self.minimum, self.maximum = self.maximum, self.minimum
        elif self.minimum == self.maximum:
            raise GraphicsError("\n\nMinimum and Maximum value cannot be the same")

        if orientation == "vertical":
            self.p1 = Point(self.anchor.x, self.anchor.y - length / 2)
            self.p2 = Point(self.anchor.x, self.anchor.y + length / 2)

            self.scale = self.p1.distance_y(self.p2) / (self.maximum - self.minimum)
            self.value_points = [y * self.scale + self.p1.y for y in range(int((maximum - minimum + 1) / step))]
        else:
            self.p1 = Point(self.anchor.x - length / 2, self.anchor.y)
            self.p2 = Point(self.anchor.x + length / 2, self.anchor.y)

            self.scale = self.p1.distance_x(self.p2) / (self.maximum - self.minimum)
            self.value_points = [x * self.scale + self.p1.x for x in range(int((maximum - minimum + 1) / step))]

        self.graphwin = None
        self.drawn = False
        self.selected = False

        self.orientation = orientation
        self.length = length
        self.step = step

        self.pointer = pointer
        self.text = text
        self.line = None

        self.style = style
        self.outline = outline
        self.width = width

        self.style = style
        self.label = label

        self.set_value(value)
        self.reload()

        GraphicsObject.slider_instances.add(self)

    def __repr__(self):
        return "SlideBar({}, {}, {}-{})".format(self.p1, self.p2, self.minimum, self.maximum)

    def set_selected(self, selected=True):
        self.selected = selected

    def draw(self, canvas):
        self.graphwin = canvas

        if self.text is not None:
            self.text.draw(canvas)

        self.line.draw(canvas)
        self.pointer.draw(canvas)

        self.drawn = True
        return self

    def get_anchor(self):
        return self.anchor

    def reload(self):
        self.line = Line(self.p1, self.p2, style=self.style, outline=self.outline, outline_width=self.width)

        if self.text is not None:
            self.text.set_text(f"{self.label}{self.state}")

        try:
            selected_clicks = self.pointer.selected_clicks
        except AttributeError:
            selected_clicks = -1

        if self.orientation == "vertical":
            self.pointer.move_to(self.anchor.x, ((self.state - self.minimum) * self.scale) + self.p1.y)
        else:
            self.pointer.move_to(((self.state - self.minimum) * self.scale) + self.p1.x, self.anchor.y)

        self.pointer.selected_clicks = selected_clicks

    def undraw(self):
        self.line.undraw()

        if self.text is not None:
            self.text.undraw()
        self.pointer.undraw()
        self.drawn = False

    def redraw(self):
        self.undraw()
        self.reload()
        self.draw(self.graphwin)

    def set_value(self, value):
        if value < self.minimum or value > self.maximum:
            raise GraphicsError(
                f"\n\nValue to set the Slider Bar must be within {self.minimum} and {self.maximum} not {value}")
        self.state = value
        try:
            self.redraw()
        except AttributeError:
            self.reload()

    def change_value(self, value):
        self.set_value(self.state + value)

    def mouse_value(self, mouse_pos):
        self.set_value(self.minimum + self.value_points.index(min(self.value_points, key=lambda x: abs(x - mouse_pos)))
                       * self.step)

    def move(self, dx, dy):
        self.undraw()

        self.line.move(dx, dy)
        if self.text is not None:
            self.text.move(dx, dy)
        self.pointer.move(dx, dy)

        self.p1 = Point(self.p1.x + dx, self.p1.y + dy)
        self.p2 = Point(self.p2.x + dx, self.p2.y + dy)

        self.draw(self.graphwin)

    def is_clicked(self, mouse_pos):
        return self.pointer.is_clicked(mouse_pos) or self.line.is_clicked(mouse_pos)

    def is_selected(self):
        return self.pointer.is_selected()

    def get_value(self):
        return self.state

    def bind_to_mouse_wheel(self, state=True):
        if state:
            if self not in GraphicsObject.slider_instances_bound:
                GraphicsObject.slider_instances_bound.add(self)
        else:
            GraphicsObject.slider_instances_bound.discard(self)



from goopylib.styles import *
from goopylib.util import *
from goopylib.constants import *

from goopylib.colours import *

from goopylib.objects.imports import *
from goopylib.math.imports import *
from goopylib.sound.imports import *

import time



from goopylib.objects.GraphicsObject import GraphicsObject
from goopylib.objects.Image import Image
from goopylib.util import GraphicsError

from os import listdir as oslistdir
from os.path import dirname as osdirname
from os.path import isdir as osisdir
from time import time as timetime


class AnimatedImage(GraphicsObject):
    def __init__(self, p, frames, align="center", cursor="arrow", update_time=1/30, layer=0,
                 tag=None, autoflush=True, frame_callback=None, bounds=None):

        if not (callable(frame_callback) or frame_callback is None):
            raise GraphicsError("\n\nGraphicsError: Frame Increment callback for Animated Image must be a function, "
                                f"not {frame_callback}")

        self.anchor = p.copy()

        self.frame = 0
        self.number_of_frames = len(frames)
        self.drawn_frame = 0
        self.frame_bound_objects = {*()}

        self.imgs = [Image(p, path, align=align, cursor=cursor, bounds=bounds) for path in frames]

        self.last_update_time = timetime()
        self.update_time = update_time
        self.autoflush = autoflush

        self.moved = False

        GraphicsObject.__init__(self, [], cursor=cursor, layer=layer, tag=tag)

        self.callbacks["frame increment"] = frame_callback

    # -------------------------------------------------------------------------
    # INTERNAL FUNCTIONS

    def __repr__(self):
        return f"AnimatedImage({self.anchor}, {self.imgs[self.frame]} and {self.number_of_frames - 1} others.)"
    
    def __iter__(self):
        for img in self.imgs:
            yield img

    def _draw(self, canvas, options):
        self.imgs[self.frame].anchor = self.anchor
        self.imgs[self.frame].draw(canvas, _internal_call=True)

        self.drawn_frame = self.frame
        if self.graphwin != canvas:
            for img in self.imgs:
                img.graphwin = canvas

        if self.autoflush and self not in GraphicsObject.animated_image_instances:
            GraphicsObject.animated_image_instances.add(self)

    def _move(self, dx, dy):
        self.anchor[0] += dx
        self.anchor[1] += dy

    def _rotate(self, dr, sampling=None, center=None):
        for img in self.imgs:
            img.rotate(dr)

    def destroy(self):
        GraphicsObject.objects.discard(self)
        GraphicsObject.draggable_objects.discard(self)
        GraphicsObject.cursor_objects.discard(self)

        if self in GraphicsObject.redraw_on_frame[self.layer]:
            GraphicsObject.redraw_on_frame[self.layer].remove(self)

        for img in self.imgs:
            img.destroy()

        self.drawn = False
        self.graphwin = None
        self.id = None

        return self
            
    def move_up_layer(self, layers=1):
        if not isinstance(layers, int):
            raise GraphicsError(f"\n\nGraphicsError: layers to move up must be an integer, not {layers}")
        if layers < 0:
            raise GraphicsError("\n\nGraphicsError: layers to move up must be greater than (or equal to) 0, "
                                f"not {layers}")

        GraphicsObject.object_layers[self.layer].remove(self)
        self.layer += layers

        while self.layer > len(GraphicsObject.object_layers) - 1:
            GraphicsObject.object_layers.append([])
        GraphicsObject.object_layers[self.layer].add(self)

        for obj in self.imgs:
            obj.move_up_layer(layers=layers)
        return self

    def move_down_layer(self, layers=1):
        if not isinstance(layers, int):
            raise GraphicsError(f"\n\nGraphicsError: layers to move down must be an integer, not {layers}")
        if layers < 0:
            raise GraphicsError("\n\nGraphicsError: layers to move down must be greater than (or equal to) 0, "
                                f"not {layers}")

        GraphicsObject.object_layers[self.layer].remove(self)
        self.layer += layers
        if self.layer < 0:
            self.layer = 0

        while self.layer > len(GraphicsObject.object_layers) - 1:
            GraphicsObject.object_layers.append([])
        GraphicsObject.object_layers[self.layer].add(self)

        for obj in self.imgs:
            obj.move_down_layer(layers=layers)
        return self

    def set_layer(self, layer=0):
        if not isinstance(layer, int):
            raise GraphicsError(f"\n\nGraphicsError: layer to set to must be an integer, not {layer}")
        if layer < 0:
            raise GraphicsError("\n\nGraphicsError: layer to set to must be greater than (or equal to) 0, "
                                f"not {layer}")

        for obj in self.imgs:
            obj.set_layer(layer=layer)

        GraphicsObject.object_layers[self.layer].remove(self)
        while layer > len(GraphicsObject.object_layers) - 1:
            GraphicsObject.object_layers.append({*()})
        GraphicsObject.object_layers[layer].add(self)

        self.layer = layer

        return self

    def set_autoflush(self, autoflush):
        self.autoflush = autoflush
        if self.autoflush:
            if self.drawn and self not in GraphicsObject.animated_image_instances:
                GraphicsObject.animated_image_instances.add(self)

        elif self in GraphicsObject.animated_image_instances:
            GraphicsObject.animated_image_instances.remove(self)

    def increment_frame(self, _time=None, _internal_call=False):
        self.frame += 1
        if self.frame == self.number_of_frames:
            self.frame = 0

        if _time is None:
            self.last_update_time = timetime()
        else:
            self.last_update_time = _time

        for obj in self.frame_bound_objects:
            if obj.drawn:
                obj.increment_frame(_time=_time, _internal_call=_internal_call)

        if self.callbacks["frame increment"] is not None:
            self.callbacks["frame increment"]()

        self._update_layer()
        return self

    def set_frame_increment_callback(self, func):
        if not (callable(func) or func is None):
            raise GraphicsError("\n\nGraphicsError: Frame Increment callback for Animated Image must be a function, "
                                f"not {func}")
        self.callbacks["frame increment"] = func
        return self

    def decrement_frame(self, _time=None, _internal_call=False):
        self.frame -= 1
        if self.frame == -1:
            self.frame = self.number_of_frames - 1

        if _time is None:
            self.last_update_time = timetime()
        else:
            self.last_update_time = _time

        for obj in self.frame_bound_objects:
            obj.decrement_frame(_time=_time, _internal_call=_internal_call)
        self._update_layer()
        return self

    def set_frame(self, frame, _time=None, _internal_call=False):

        if not (0 <= frame < self.number_of_frames):
            raise GraphicsError("\n\nGraphicsError: Frame to set the Animated Image to must be between (inclusive) 0 & "
                                f"{self.number_of_frames - 1}, not {frame}")

        self.frame = frame

        if _time is None:
            self.last_update_time = timetime()
        else:
            self.last_update_time = _time
        self._update_layer()
        return self

    def bind_frame_to(self, obj):
        if not isinstance(obj, AnimatedImage):
            if obj in GraphicsObject.tagged_objects:
                obj = GraphicsObject.tagged_objects[obj]
                if not isinstance(obj, AnimatedImage):
                    raise GraphicsError("\n\nGraphicsError: object to bind frame to must be an Animated Image object, "
                                        f"not {obj}")
            else:
                raise GraphicsError("\n\nGraphicsError: object to bind frame to must be an Animated Image object, "
                                    f"not {obj}")
        obj.frame_bound_objects.add(self)
        self.set_autoflush(False)
        return self

    # -------------------------------------------------------------------------
    # OTHER & IMPORTANT FUNCTIONS

    def _undraw(self, set_blinking=True):
        self.imgs[self.drawn_frame].undraw(set_blinking=set_blinking)

        if self in GraphicsObject.animated_image_instances:
            GraphicsObject.animated_image_instances.remove(self)

    def base_undraw(self):
        for img in self.imgs:
            if img.drawn:
                img.base_undraw()

    def is_clicked(self, mouse_pos):
        return self.imgs[self.frame].is_clicked(mouse_pos)

    def save(self, filename):
        self.imgs[self.frame].save(filename)
        return self

    # -------------------------------------------------------------------------
    # IMAGE COLOUR & ENHANCE FUNCTIONS

    def set_contrast(self, level):
        for img in self.imgs:
            img.set_contrast(level)
        return self

    def change_contrast(self, level):
        for img in self.imgs:
            img.change_contrast(level)
        return self

    def reset_contrast(self):
        self.set_contrast(0)

    # -------------------------------------------------------------------------
    # IMAGE MANIPULATION FUNCTIONS

    def crop(self, left=0, top=0, right=500, down=500, align="center"):
        for img in self.imgs:
            img.crop(left=left, top=top, right=right, down=down, align=align)
        return self

    # Blending & Compositing Functions

    def blend(self, img, alpha):
        for img in self.imgs:
            img.blend(img, alpha)
        return self

    def alpha_composite(self, other):
        for img in self.imgs:
            img.alpha_composite(other)
        return self

    def composite(self, other, img_mask):
        for img in self.imgs:
            img.composite(other, img_mask)
        return self

    def convert_greyscale(self):
        for img in self.imgs:
            img.convert_greyscale()
        return self

    def convert_binary(self):
        for img in self.imgs:
            img.convert_binary()
        return self

    # ----------------------------------
    # Flipping Functions

    def flip(self, x_axis=True, y_axis=True):
        for img in self.imgs:
            img.flip(x_axis=x_axis, y_axis=y_axis)
        return self

    def flip_x(self):
        for img in self.imgs:
            img.flip_x()
        return self

    def flip_y(self):
        for img in self.imgs:
            img.flip_y()
        return self

    def flip_xy(self):
        for img in self.imgs:
            img.flip_xy()
        return self

    def transverse(self):
        for img in self.imgs:
            img.transverse()
        return self

    def transpose(self):
        for img in self.imgs:
            img.transpose()
        return self

    # ----------------------------------
    # Skew Transformation Functions

    def skew_x(self, scale=0.3, sampling="bicubic", align="center"):
        for img in self.imgs:
            img.skew_x(scale=scale, sampling=sampling, align=align)
        return self

    def skew_y(self, scale=0.3, sampling=None, align="center"):
        for img in self.imgs:
            img.skew_y(scale=scale, sampling=sampling, align=align)
        return self

    def skew_xy(self, x_scale=0.3, y_scale=None, sampling=None, x_align="center", y_align=None):
        for img in self.imgs:
            img.skew_xy(x_scale=x_scale, y_scale=y_scale, sampling=sampling, x_align=x_align, y_align=y_align)
        return self

    # ----------------------------------
    # Resizing Functions

    def resize(self, width, height, sampling=None, _external_call=True):
        for img in self.imgs:
            img.resize(width, height, sampling=sampling, _external_call=_external_call)
        return self

    def resize_height(self, height, preserve_aspect_ratio=False, sampling=None, _external_call=True):
        for img in self.imgs:
            img.resize_height(height, preserve_aspect_ratio=preserve_aspect_ratio, sampling=sampling,
                                _external_call=_external_call)
        return self

    def resize_width(self, width, preserve_aspect_ratio=False, sampling=None, _external_call=True):
        for img in self.imgs:
            img.resize_width(width, preserve_aspect_ratio=preserve_aspect_ratio, sampling=sampling,
                               _external_call=_external_call)
        return self

    def resize_factor(self, factor, sampling=None, _external_call=True):
        for img in self.imgs:
            img.resize_factor(factor, sampling=sampling, _external_call=_external_call)
        return self

    def resize_width_factor(self, factor, sampling=None, _external_call=True):
        for img in self.imgs:
            img.resize_width_factor(factor, sampling=sampling, _external_call=_external_call)
        return self

    def resize_height_factor(self, factor, sampling=None, _external_call=True):
        for img in self.imgs:
            img.resize_width_factor(factor, sampling=sampling, _external_call=_external_call)
        return self

    def resize_to_fit(self, obj, sampling=None, _external_call=True):
        for img in self.imgs:
            img.resize_to_fit(obj, sampling=sampling, _external_call=_external_call)
        return self

    def resize_to_fit_width(self, obj, preserve_aspect_ratio=False, sampling=None, _external_call=True):
        for img in self.imgs:
            img.resize_to_fit_width(obj, preserve_aspect_ratio=preserve_aspect_ratio, sampling=sampling,
                                      _external_call=_external_call)
        return self

    def resize_to_fit_height(self, obj, preserve_aspect_ratio=False, sampling=None, _external_call=True):
        for img in self.imgs:
            img.resize_to_fit_height(obj, preserve_aspect_ratio=preserve_aspect_ratio, sampling=sampling,
                                       _external_call=_external_call)
        return self

    # ----------------------------------
    # Blurring & Sharpening Functions

    def blur(self):
        for img in self.imgs:
            img.blur()
        return self

    def blur_box(self, radius=3):
        for img in self.imgs:
            img.blur_box(radius=radius)
        return self

    def blur_gaussian(self, radius=3):
        for img in self.imgs:
            img.blur_gaussian(radius=radius)
        return self

    def sharpen(self, radius=3, percent=150):
        for img in self.imgs:
            img.sharpen(radius=radius, percent=percent)
        return self

    # ----------------------------------
    # Filter Functions

    def filter_contour(self):
        for img in self.imgs:
            img.filter_contour()
        return self

    def filter_detail(self):
        for img in self.imgs:
            img.filter_detail()
        return self

    def filter_emboss(self):
        for img in self.imgs:
            img.filter_emboss()
        return self

    def filter_find_edges(self):
        for img in self.imgs:
            img.filter_find_edges()
        return self

    def filter_sharpen(self):
        for img in self.imgs:
            img.filter_sharpen()
        return self

    def filter_smooth(self):
        for img in self.imgs:
            img.filter_smooth()
        return self

    def filter_more_smooth(self):
        for img in self.imgs:
            img.filter_more_smooth()
        return self

    def filter_enhance_edge(self):
        for img in self.imgs:
            img.filter_enhance_edge()
        return self

    def filter_more_enhance_edge(self):
        for img in self.imgs:
            img.filter_more_enhance_edge()
        return self

    # -------------------------------------------------------------------------
    # GETTER FUNCTIONS

    def get_anchor(self):
        return self.anchor

    def get_width(self):
        return self.imgs[0].get_width()

    def get_height(self):
        return self.imgs[0].get_height()

    def get_aspect_ratio(self):
        return self.imgs[0].get_aspect_ratio()

    def get_pixel(self, x, y):
        return self.imgs[0].get_pixel(x, y)

    def get_format(self):
        return self.imgs[0].get_format()

    def get_mode(self):
        return self.imgs[0].get_mode()

    def get_contrast(self):
        return self.imgs[0].get_contrast()

    def get_frame(self):
        return self.frame

    def get_number_of_frames(self):
        return len(self.imgs)

    def get_current_frame(self):
        return self.imgs[self.frame]

    def get_all_frames(self):
        return self.imgs

    # -------------------------------------------------------------------------
    # SETTER FUNCTIONS

    def set_pixel(self, x, y, colour):
        for img in self.imgs:
            img.set_pixel(x, y, colour)
        return self



from goopylib.objects.Checkbox import Checkbox
from goopylib.objects.GraphicsObject import GraphicsObject

class Button(GraphicsObject):
    def __init__(self, graphic, hover_graphic=None, clicked_graphic=None, disabled_graphic=None, bounds=None,
                 disable=False, autoflush=True, label=None, layer=0, tag=None):

        self.disabled_graphic_given = True
        self.clicked_graphic_given = True
        self.hover_graphic_given = True

        self.normal_graphic = graphic
        if hover_graphic is not None:
            self.hover_graphic = hover_graphic
        else:
            self.hover_graphic = graphic
            self.hover_graphic_given = False

        if clicked_graphic is not None:
            self.clicked_graphic = clicked_graphic
        else:
            self.clicked_graphic = graphic
            self.clicked_graphic_given = False

        if disabled_graphic is not None:
            self.disabled_graphic = disabled_graphic
        else:
            self.disabled_graphic = graphic
            self.disabled_graphic_given = False

        self.graphic = self.normal_graphic
        self.anchor = self.graphic.anchor
        self.drawn_graphic = self.graphic

        self.label = label
        self.is_object_disabled = disable

        if not isinstance(self, Checkbox):
            GraphicsObject.button_instances.add(self)

        GraphicsObject.__init__(self, options=(), layer=layer, tag=tag, bounds=bounds)

    def __repr__(self):
        return f"Button({self.graphic}, {self.normal_graphic}, {self.hover_graphic})"

    def _draw(self, canvas, options):
        self.graphic.draw(canvas, _internal_call=True)
        self.drawn_graphic = self.graphic

        self.anchor = self.graphic.anchor

        if self.label is not None:
            self.label.draw(canvas, _internal_call=True)

    def destroy(self):
        GraphicsObject.objects.discard(self)
        GraphicsObject.object_layers[self.layer].discard(self)
        GraphicsObject.draggable_objects.discard(self)
        GraphicsObject.cursor_objects.discard(self)
        GraphicsObject.button_instances.discard(self)

        if self.hover_graphic_given:
            self.hover_graphic.destroy()
        self.normal_graphic.destroy()
        if self.clicked_graphic_given:
            self.clicked_graphic.destroy()
        if self.disabled_graphic_given:
            self.disabled_graphic.destroy()

        if self in GraphicsObject.redraw_on_frame[self.layer]:
            GraphicsObject.redraw_on_frame[self.layer].remove(self)

        if self.label is not None:
            self.label.destroy()

        self.drawn = False
        self.graphwin = None
        self.id = None

        return self

    def _move(self, dx, dy):
        if self.hover_graphic_given:
            self.hover_graphic._move(dx, dy)
        self.normal_graphic._move(dx, dy)
        if self.clicked_graphic_given:
            self.clicked_graphic._move(dx, dy)
        if self.disabled_graphic_given:
            self.disabled_graphic._move(dx, dy)

        if self.label is not None:
            self.label._move(dx, dy)

    def _rotate(self, dr):
        self.disabled_graphic.rotate(dr)
        self.normal_graphic.rotate(dr)
        self.hover_graphic.rotate(dr)
        self.clicked_graphic.rotate(dr)
        if self.label is not None:
            self.label.rotate(dr)

    def _undraw(self, set_blinking=False):
        self.drawn_graphic.undraw()
        if self.label is not None:
            self.label.undraw()
        return self

    def base_undraw(self):
        self.drawn_graphic.undraw()
        if self.label is not None:
            self.label.undraw()

    def get_width(self):
        return self.graphic.get_width()

    def get_height(self):
        return self.graphic.get_height()

    def set_graphic(self, graphic):
        self.undraw()
        if graphic == "Normal":
            self.graphic = self.normal_graphic
        elif graphic == "Hover":
            self.graphic = self.hover_graphic
        elif graphic == "Clicked":
            self.graphic = self.clicked_graphic
        elif graphic == "Disabled":
            self.graphic = self.disabled_graphic

        return self.draw(self.graphwin)

    def is_clicked(self, mouse_pos):
        if self.bounds is None:
            return self.graphic.is_clicked(mouse_pos) and not self.is_object_disabled
        else:
            return self.bounds.is_clicked(mouse_pos)

    def get_anchor(self):
        return self.graphic.anchor

    def disable(self):
        self.is_object_disabled = True

        self.undraw()
        self.graphic = self.disabled_graphic
        return self.draw(self.graphwin)

    def enable(self):
        self.is_object_disabled = False

        self.undraw()
        self.graphic = self.normal_graphic
        return self.draw(self.graphwin)

    def is_enabled(self):
        return not self.is_object_disabled

    def is_disabled(self):
        return self.is_object_disabled

    def toggle_enabled(self):
        return self.set_enabled(not self.is_object_disabled)

    def set_enabled(self, enable):
        if enable:
            return self.enable()
        else:
            return self.disable()

    def change_graphic(self, graphic=0, hover_graphic=0, clicked_graphic=0, disabled_graphic=0, label=0):
        self.undraw()
        if hover_graphic != 0:
            self.hover_graphic = hover_graphic
        if graphic != 0:
            self.normal_graphic = graphic
        if clicked_graphic != 0:
            self.clicked_graphic = clicked_graphic
        if disabled_graphic != 0:
            self.disabled_graphic = disabled_graphic

        if label != 0:
            self.label = label

        return self.draw(self.graphwin)



import goopylib.math.c_bezier_curve


def combination(n, k):
    return goopylib.math.c_bezier_curve.combination(n, k)


def bernstein_polynomial(i, n, t):
    return goopylib.math.c_bezier_curve.bernstein_polynomial(i, n, t)


def bezier_curve(t, control_points):
    return goopylib.math.c_bezier_curve.bezier_curve(t, control_points)


def rational_bezier_curve(t, control_points, weights):
    return goopylib.math.c_bezier_curve.rational_bezier_curve(t, control_points, weights)



import goopylib.objects._BBox as gpBBox
from goopylib.objects.Rectangle import Rectangle

from goopylib.util import GraphicsError

class Oval(gpBBox.BBox):

    def __init__(self, center, radius1, radius2, bounds=None, fill=None, outline=None, outline_width=None,
                 cursor="arrow", layer=0, tag=None):

        if not isinstance(center, list):
            raise GraphicsError("\n\nGraphicsError: center argument for Oval must be a list in the form [x, y], "
                                f"not {center}")

        if not isinstance(radius1, int):
            raise GraphicsError(f"\n\nGraphicsError: radius1 argument for Oval must be an integer, not {radius1}")
        if not isinstance(radius2, int):
            raise GraphicsError(f"\n\nGraphicsError: radius2 argument for Oval must be an integer, not {radius2}")

        p1 = [center[0] - radius1, center[1] - radius2]
        p2 = [center[0] + radius1, center[1] + radius2]

        self.radius1 = radius1
        self.radius2 = radius2
        
        self.center = center

        gpBBox.BBox.__init__(self, p1, p2, bounds=bounds, fill=fill, outline=outline, outline_width=outline_width,
                             cursor=cursor, layer=layer, tag=tag)

    def __repr__(self):
        return f"Oval(anchor={self.center}, radius1={self.radius1}, radius2={self.radius2})"

    def clone(self, new_tag=None):
        other = Oval(self.center, self.radius1, self.radius1, bounds=self.bounds.clone(), fill=self.fill,
                     outline=self.outline, outline_width=self.outline_width, cursor=self.cursor, layer=self.layer,
                     tag=new_tag)
        return other

    def _draw(self, canvas, options):
        x1, y1 = canvas.to_screen(self.p1[0], self.p1[1])
        x2, y2 = canvas.to_screen(self.p2[0], self.p2[1])

        return canvas.create_oval(x1, y1, x2, y2, fill=self.fill, outline=self.outline, width=self.outline_width)

    def _set_resizable(self, resizables, top_bounds=None, bottom_bounds=None, left_bounds=None, right_bounds=None,
                       thickness=10):

        # resizables is a list [top, bottom, left, right] where each of its elements are Booleans defining whether the
        # rectangle is resizable in that direction

        self.resizing_bounds = {}  # Stores the bounds of the resizing objects for 'top', 'bottom', 'left', & 'right'
        if resizables[0]:  # Checks if the object is set to be resized from the top
            if top_bounds is None:  # The user hasn't supplied a custom bound
                self.resizing_bounds["top"] = Rectangle([self.p1[0] - thickness / 2, self.p1[1] - thickness / 2],
                                                        [self.p2[0] + thickness / 2, self.p1[1] + thickness / 2],
                                                        cursor="vertical-resizing-arrow")
            elif isinstance(top_bounds, GraphicsObject):  # The user supplied a GraphicsObject as the bound
                self.resizing_bounds["top"] = top_bounds

            elif top_bounds in GraphicsObject.tagged_objects:  # The user supplied a tag referencing a GraphicsObject
                self.resizing_bounds["top"] = GraphicsObject.tagged_objects[top_bounds]

            else:  # The user provided an invalid argument, this raises an error
                raise GraphicsObject("\n\nGraphicsError: Top bound for Rectangle resizing must be a None, "
                                     f"GraphicsObject or tag referencing a GraphicsObject, not {top_bounds}")

        if resizables[1]:  # Checks if the object is set to be resized from the bottom
            if bottom_bounds is None:  # The user hasn't supplied a custom bound
                self.resizing_bounds["bottom"] = Rectangle([self.p2[0] + thickness / 2, self.p2[1] - thickness / 2],
                                                           [self.p1[0] - thickness / 2, self.p2[1] + thickness / 2],
                                                           cursor="vertical-resizing-arrow")
            elif isinstance(bottom_bounds, GraphicsObject):  # The user supplied a GraphicsObject as the bound
                self.resizing_bounds["bottom"] = top_bounds

            elif bottom_bounds in GraphicsObject.tagged_objects:  # The user supplied a tag referencing a GraphicsObject
                self.resizing_bounds["bottom"] = GraphicsObject.tagged_objects[top_bounds]

            else:  # The user provided an invalid argument, this raises an error
                raise GraphicsObject("\n\nGraphicsError: Top bound for Rectangle resizing must be a None, "
                                     f"GraphicsObject or tag referencing a GraphicsObject, not {top_bounds}")

        if resizables[2]:  # Checks if the object is set to be resized from the left
            if left_bounds is None:  # The user hasn't supplied a custom bound
                self.resizing_bounds["left"] = Rectangle([self.p1[0] - thickness / 2, self.p1[1] - thickness / 2],
                                                         [self.p1[0] + thickness / 2, self.p2[1] + thickness / 2],
                                                         cursor="horizontal-resizing-arrow")
            elif isinstance(left_bounds, GraphicsObject):  # The user supplied a GraphicsObject as the bound
                self.resizing_bounds["left"] = left_bounds

            elif left_bounds in GraphicsObject.tagged_objects:  # The user supplied a tag referencing a GraphicsObject
                self.resizing_bounds["left"] = GraphicsObject.tagged_objects[left_bounds]

            else:  # The user provided an invalid argument, this raises an error
                raise GraphicsObject("\n\nGraphicsError: Top bound for Rectangle resizing must be a None, "
                                     f"GraphicsObject or tag referencing a GraphicsObject, not {top_bounds}")

        if resizables[3]:  # Checks if the object is set to be resized from the right
            if right_bounds is None:  # The user hasn't supplied a custom bound
                self.resizing_bounds["right"] = Rectangle([self.p2[0] + thickness / 2, self.p2[1] + thickness / 2],
                                                          [self.p2[0] - thickness / 2, self.p1[1] - thickness / 2],
                                                          cursor="horizontal-resizing-arrow")
            elif isinstance(right_bounds, GraphicsObject):  # The user supplied a GraphicsObject as the bound
                self.resizing_bounds["right"] = right_bounds

            elif right_bounds in GraphicsObject.tagged_objects:  # The user supplied a tag referencing a GraphicsObject
                self.resizing_bounds["right"] = GraphicsObject.tagged_objects[right_bounds]

            else:  # The user provided an invalid argument, this raises an error
                raise GraphicsObject("\n\nGraphicsError: Top bound for Rectangle resizing must be a None, "
                                     f"GraphicsObject or tag referencing a GraphicsObject, not {top_bounds}")

    def is_clicked(self, mouse_pos):
        if mouse_pos is None:
            return False

        if self.bounds is None:
            return (((mouse_pos[0] - self.anchor[0]) ** 2) / self.radius1 ** 2) \
                   + (((mouse_pos[1] - self.anchor[1]) ** 2) / self.radius2 ** 2) < 1
        else:
            return self.bounds.is_clicked(mouse_pos)


from goopylib.objects.GraphicsObject import GraphicsObject



import math

# References for this: https://www.youtube.com/watch?v=qhQrRCJ-mVg

def py_bezier_curve(t, control_points):
    size = len(control_points)
    if size > 5:
        sum_x, sum_y = 0, 0

        size = len(control_points) - 1
        for i in range(0, size + 1):
            coeff = py_bernstein_polynomial(i, size, t)
            sum_y += coeff * control_points[i][1]
            sum_x += coeff * control_points[i][0]

        return sum_x, sum_y

    elif size == 2:
        return ((1 - t) * control_points[0][0]) + (t * control_points[1][0]), \
               ((1 - t) * control_points[0][1]) + (t * control_points[1][1])

    elif size == 3:
        t2 = t * t
        t *= 2

        return ((t2 - t + 1) * control_points[0][0]) + (((-2 * t2) + t) * control_points[1][0]) + \
               (t2 * control_points[2][0]), \
               ((t2 - t + 1) * control_points[0][1]) + (((-2 * t2) + t) * control_points[1][1]) + \
               (t2 * control_points[2][1])

    elif size == 4:
        t2 = t * t
        t3 = t2 * t

        t *= 3

        return ((-t3 + (3 * t2) - t + 1) * control_points[0][0]) + \
               (((3 * t3) - (6 * t2) + t) * control_points[1][0]) + \
               (((-3 * t3) + (3 * t2)) * control_points[2][0]) + \
               (t3 * control_points[3][0]), \
               ((-t3 + (3 * t2) - t + 1) * control_points[0][1]) + \
               (((3 * t3) - (6 * t2) + t) * control_points[1][1]) + \
               (((-3 * t3) + (3 * t2)) * control_points[2][1]) + (t3 * control_points[3][1])

    elif size == 5:
        t2 = t * t
        t3 = t2 * t
        t4 = t3 * t
        t *= 4

        return ((t4 - (4 * t3) + (6 * t2) - t + 1) * control_points[0][0]) + \
               (((-4 * t4) + (12 * t3) - (12 * t2) + t) * control_points[1][0]) + \
               (((6 * t4) - (12 * t3) + (6 * t2)) * control_points[2][0]) + \
               (((-4 * t4) + (4 * t3)) * control_points[3][0]) + \
               (t4 * control_points[4][0]), \
               ((t4 - (4 * t3) + (6 * t2) - t + 1) * control_points[0][1]) + \
               (((-4 * t4) + (12 * t3) - (12 * t2) + t) * control_points[1][1]) + \
               (((6 * t4) - (12 * t3) + (6 * t2)) * control_points[2][1]) + \
               (((-4 * t4) + (4 * t3)) * control_points[3][1]) + \
               (t4 * control_points[4][1])

    return 0, 0

def py_rational_bezier_curve(t, control_points, weights):
    sum_x_numerator, sum_y_numerator = 0, 0
    sum_x, sum_y = 0, 0

    degree = len(control_points) - 1

    for i in range(0, len(control_points)):
        coeff = py_bernstein_polynomial(i, degree, t) * weights[i]
        sum_x_numerator += coeff * control_points[i][0]
        sum_x += coeff

        sum_y_numerator += coeff * control_points[i][1]
        sum_y += coeff

    return sum_x_numerator / sum_x, sum_y_numerator / sum_y


def py_bernstein_polynomial(i, n, t):
    return py_combination(n, i) * (t ** i) * ((1 - t) ** (n - i))


def py_combination(n, k):
    return math.factorial(n) / (math.factorial(k) * math.factorial(n - k))



from tkinter import Canvas as tkCanvas
from tkinter import Toplevel as tkToplevel

from PIL import ImageGrab  # Required to take snapshots of the GraphWin to save as an Image

import warnings
from os.path import isfile as osisfile
from time import time as timetime

from goopylib.styles import *
from goopylib.util import GraphicsError, GraphicsWarning, resource_path
from goopylib.constants import _root, RELIEF, CURSORS
from goopylib.math.py_easing import *

from goopylib._internal_classes import Transform


class Window(tkCanvas):
    """A Window is a toplevel window for displaying graphics."""
    instances = []  # a list of all instances of this class

    # Most of the parameters are self explanatory, but for those that might not be:

    # Resizable Height & Width: Whether you can resize the window by pulling its sides
    # Autoflush: The window will update automatically if True
    # Style: The colour styles have attributes called 'background-colour' which the window will use if bk_colour=None
    def __init__(self, title="New Goopylib Window", width=800, height=600, min_width=1, min_height=1, max_width=None,
                 max_height=None, x_pos=0, y_pos=0, resizable_width=False, resizable_height=False, bounds=None,
                 bk_colour=None, icon=None, autoflush=False, cursor="arrow", border_relief="flat",
                 border_width=0, remove_title_bar=False):

        # Making sure all the arguments are valid and raising errors if not

        if not isinstance(title, str):
            raise GraphicsError(f"\n\nThe window's title must be a string, not {title}")

        if not isinstance(width, int):
            raise GraphicsError(f"\n\nThe window's width must be an integer, not {width}")
        if width < 1:
            raise GraphicsError(f"\n\nThe window's width must be greater than 0, not {width}")
        if not isinstance(height, int):
            raise GraphicsError(f"\n\nThe window's height must be an integer, not {height}")
        if height < 1:
            raise GraphicsError(f"\n\nThe window's height must be greater than 0, not {height}")

        if not isinstance(x_pos, int):
            raise GraphicsError(f"\n\nThe window x-position (x_pos) must be an integer, not {x_pos}")
        if not isinstance(y_pos, int):
            raise GraphicsError(f"\n\nThe window y-position (y_pos) must be an integer, not {y_pos}")

        if not (isinstance(bk_colour, Colour) or bk_colour is None):
            raise GraphicsError("The window's background colour (bk_colour) must be None or a Colour Objector, "
                                f"not {bk_colour}")
        if icon is not None:
            if not isinstance(icon, str):
                raise GraphicsError(f"The window icon must be a string (path to .ico texture) or None, not {icon}")
            if not icon.endswith(".ico"):
                raise GraphicsError("The icon file must be a .ico type. "
                                    "You can use an online converter to convert your file")

        if not isinstance(min_width, int):
            raise GraphicsError(f"\n\nThe window's minimum width (min_width) must be an integer, not {min_width}")
        if min_width < 1:
            raise GraphicsError(f"\n\nThe window's minimum width must be greater than 0, not {min_width}")

        if not isinstance(min_height, int):
            raise GraphicsError(f"\n\nThe window's minimum height (min_height) must be an integer, not {min_height}")
        if min_height < 1:
            raise GraphicsError(f"\n\nThe window's minimum height must be greater than 0, not {min_height}")

        if max_width is not None:
            if isinstance(max_width, int):
                if min_width > max_width:
                    raise GraphicsError(
                        f"Window's minimum width ({min_width}) is larger than its maximum width ({max_width})")

                if width > max_width:
                    warning = f"\n\nWindow width ({width}) is greater than window's maximum width ({max_width}). " \
                              f"Window maximum width has been set to {width}."
                    max_width = width
                    warnings.warn(warning, GraphicsWarning)

            else:
                raise GraphicsError(f"\n\nThe window's maximum width (max_width) must be an integer, not {max_width}")

        if max_height is not None:
            if isinstance(max_height, int):
                if min_height > max_height:
                    raise GraphicsError(f"Window's minimum height ({min_height}) is larger than its maximum height "
                                        f"({max_height})'")

                if height > max_height:
                    warning = f"\n\nWindow height ({height}) is greater than window's maximum height ({max_height}). " \
                              f"Window maximum height has been set to {height}."
                    max_height = height
                    warnings.warn(warning, GraphicsWarning)

            else:
                raise GraphicsError("\n\nThe window's maximum height (max_height) must be an integer, "
                                    f"not {max_height}")

        if not isinstance(cursor, str):
            raise GraphicsError(f"\n\nThe cursor must be a string, not {cursor}")
        if not cursor.lower() in CURSORS.keys():
            raise GraphicsError(f"\n\nThe cursor for the window must be one of {list(CURSORS.keys())}, not {cursor}")

        if not isinstance(border_relief, str):
            raise GraphicsError(f"\n\nThe border relief (border_relief) must be a string, not {border_relief}")
        if not border_relief.lower() in RELIEF:
            raise GraphicsError(f"\n\nThe relief for the window border must be one of {RELIEF}, not {border_relief}")
        if not isinstance(border_width, int):
            raise GraphicsError(f"\n\nThe window's border width must be an integer, not {border_width}")

        if bounds is not None and not isinstance(bounds, GraphicsObject):
            raise GraphicsError(f"\n\nThe window's bounds must be a GraphicsObject, not {bounds}")

        # Making sure that the parameters provided for the dimensions of the window don't contradict with each other.
        # If they do, this raises a warning and sets the dimensions to the correct value

        if width < min_width:
            warning = f"\n\nWindow width ({width}) is less than window's minimum width ({min_width}). " \
                      f"Window minimum width has been set to {width}."
            min_width = width
            warnings.warn(warning, GraphicsWarning)

        if height < min_height:
            warning = f"\n\nWindow height ({height}) is less than window's minimum height ({min_height}). " \
                      f"Window minimum height has been set to {height}."
            min_height = height
            warnings.warn(warning, GraphicsWarning)

        master = tkToplevel(_root)  # Actually creating the window
        master.protocol("WM_DELETE_WINDOW", self.close)  # Close the window if the X button is pressed

        # Assigning a background colour, this goes through multiple checks to assign a possible value:
        #   1. The user has directly provided a colour
        #   2. The colour provided is a colour tag in the window's style
        #   3. If not, choose the background colour tag of the window's style
        #   4. If this does not exist, choose the default background colour
        if isinstance(bk_colour, Colour):
            self.bk_colour = bk_colour
        else:
            self.bk_colour = STYLES["default"]["background"]

        # Setting most of the attributes of the window
        # We set the width & height of the window to be the screen size to get rid of a bug with resizing,
        # then setting the window coordinates

        tkCanvas.__init__(self, master, width=width, height=height,
                          highlightthickness=0, bd=border_width,
                          bg=self.bk_colour, cursor=CURSORS[cursor.lower()], relief=border_relief)
        self.master.config(width=width, height=height)  # Changing the dimensions to be correct

        self.master.title(title)  # Setting the title of the widow
        self.title = title

        self.master.geometry('%dx%d+%d+%d' % (width, height, x_pos, y_pos))  # Setting the X & Y position of the window
        self.height = height
        self.width = width
        self.center = [width / 2, height / 2]

        self.x_pos = x_pos
        self.y_pos = y_pos

        self.min_width = min_width
        self.min_height = min_height
        self.master.minsize(self.min_width, self.min_height)  # Minimum size of the window

        self.max_width = max_width
        self.max_height = max_height
        self.master.maxsize(self.max_width, self.max_height)  # Maximum size of the window

        self.master.resizable(resizable_width, resizable_height)  # Is the window resizable?
        self.is_resizable = [resizable_width, resizable_height]

        self.is_gliding = False

        # The Transform Class for the window
        # This helps transform any coordinates into window coordinates (world) or absolute coords (screen)
        self.trans = None
        self.closed = False
        self.autoflush = autoflush

        self.cursor = cursor.lower()
        self.current_cursor = self.cursor

        self.border_width = border_width
        self.border_relief = border_relief.lower()

        if icon is not None:  # Setting the Icon of the Window
            if osisfile(resource_path(f"textures/{icon}")):
                self.master.iconbitmap(resource_path(f"textures/{icon}"))
            else:
                self.master.iconbitmap(resource_path(f"{icon}"))

        self.pack()

        self.imgs = 0  # Number of Images in this window
        self.items = []  # All the Graphics Objects in this window

        # Mouse Related Variables

        self.mouse_left_click = None  # These are all coordinates of these events
        self.mouse_middle_click = None
        self.mouse_right_click = None

        self.mouse_double_left_click = None
        self.mouse_double_middle_click = None
        self.mouse_double_right_click = None

        self.mouse_triple_left_click = None
        self.mouse_triple_middle_click = None
        self.mouse_triple_right_click = None

        self.mouse_left_press = None
        self.mouse_right_press = None
        self.mouse_middle_press = None

        self.total_left_mouse_clicks = 0
        self.total_middle_mouse_clicks = 0
        self.total_right_mouse_clicks = 0

        self.last_mouse_event = None
        self.mouse_pos = None

        self.left_mouse_down = False
        self.right_mouse_down = False
        self.middle_mouse_down = False
        self.mouse_in_window = False

        self.master.bind("<1>", lambda event: event.widget.focus_set())
        self.bind("<Button-1>", self._on_left_click)  # These functions are called whenever these events happen
        self.bind("<Button-2>", self._on_middle_click)
        self.bind("<Button-3>", self._on_right_click)

        self.bind("<ButtonRelease-1>", self._on_left_release)
        self.bind("<ButtonRelease-2>", self._on_middle_release)
        self.bind("<ButtonRelease-3>", self._on_right_release)

        self.bind("<ButtonPress-1>", self._on_left_press)
        self.bind("<ButtonPress-2>", self._on_middle_press)
        self.bind("<ButtonPress-3>", self._on_right_press)

        self.bind("<Motion>", self._on_mouse_motion)
        self.bind_all("<MouseWheel>", self._on_mouse_scroll)
        self.bind("<Enter>", self._on_mouse_in)
        self.bind("<Leave>", self._on_mouse_out)

        self.bind("<Double-Button-1>", self._on_left_double_click)
        self.bind("<Double-Button-2>", self._on_middle_double_click)
        self.bind("<Double-Button-3>", self._on_right_double_click)

        self.bind("<Triple-Button-1>", self._on_left_triple_click)
        self.bind("<Triple-Button-2>", self._on_middle_triple_click)
        self.bind("<Triple-Button-3>", self._on_right_triple_click)

        self._mouse_callback = None
        
        self.is_draggable_x = False
        self.is_draggable_y = False
        self.is_dragging = False
        self.bounds = bounds

        self.mouse_wheel_pos = 0
        self.mouse_scroll_amount = 0

        # Keyboard Variables

        self.last_key_pressed = None
        self.last_key_clicked = None

        self.key_pressed_with_shift = None
        self.key_pressed_with_alt = None
        self.key_pressed_with_control = None

        self.keys_down = {*()}
        self.keys_clicked = {*()}

        self.bind("<KeyPress>", self._on_key_press)
        self.bind("<Shift-KeyPress>", self._on_shift_key_press)
        self.bind("<Alt-KeyPress>", self._on_alt_key_press)
        self.bind("<Control-KeyPress>", self._on_control_key_press)

        self.bind("<KeyRelease>", self._on_key_release)

        # Animation Variables

        self.is_gliding = False  # Is the window gliding?
        self.glide_queue = []  # The next locations for the window to glide to

        master.lift()  # No idea what this does, does anyone know?
        Window.instances.append(self)

        _root.update_idletasks()
        self.set_coords(0, 0, width, height)

        if remove_title_bar:
            master.overrideredirect(True)

        self.last_update = 0

    def __repr__(self):
        if self.closed:
            return f"Closed GraphWin({self.title})"
        else:
            return f"GraphWin('{self.master.title()}', {self.master.winfo_width()}x{self.master.winfo_height()})"

    def __str__(self):
        return repr(self)

    def __check_open(self):
        if self.closed:
            raise GraphicsError("\n\nwindow is closed")

    def is_closed(self):
        return self.closed

    def is_open(self):
        return not self.closed

    # This is called automatically whenever the window needs to update
    def __autoflush(self):
        if self.autoflush:
            _root.update()

    def __set_mouse_handler(self, func):
        self._mouse_callback = func\
        
    # -------------------------------------------------------------------------
    # CALLBACK KEYBOARD FUNCTIONS

    def _on_key_release(self, e):
        if e.keysym in self.keys_down:
            self.keys_down.remove(e.keysym)
        if e.keysym not in self.keys_clicked:
            self.keys_clicked.add(e.keysym)

    def _on_key_press(self, e):
        self.last_key_pressed = e.keysym
        if e.keysym not in self.keys_down:
            self.keys_down.add(e.keysym)

    def _on_shift_key_press(self, e):
        self.key_pressed_with_shift = e.keysym

    def _on_alt_key_press(self, e):
        self.key_pressed_with_alt = e.keysym

    def _on_control_key_press(self, e):
        self.key_pressed_with_control = e.keysym

    # -------------------------------------------------------------------------
    # CALLBACK MOUSE FUNCTIONS

    # Double Click
    def _on_left_double_click(self, e):
        if self.is_open():
            self.mouse_double_left_click = e.x, e.y
            self.last_mouse_event = self.trans.world(e.x, e.y)

            if self._mouse_callback:
                self._mouse_callback([e.x, e.y])

            GraphicsObject.on_double_left_click(self)

    def _on_middle_double_click(self, e):
        if self.is_open():
            self.mouse_double_middle_click = e.x, e.y
            self.last_mouse_event = self.trans.world(e.x, e.y)

            if self._mouse_callback:
                self._mouse_callback([e.x, e.y])

            GraphicsObject.on_double_middle_click(self)

    def _on_right_double_click(self, e):
        if self.is_open():
            self.mouse_double_right_click = e.x, e.y
            self.last_mouse_event = self.trans.world(e.x, e.y)

            if self._mouse_callback:
                self._mouse_callback([e.x, e.y])

            GraphicsObject.on_double_right_click(self)

    # Triple Click

    def _on_left_triple_click(self, e):
        if self.is_open():
            self.mouse_triple_left_click = e.x, e.y
            self.last_mouse_event = self.trans.world(e.x, e.y)

            if self._mouse_callback:
                self._mouse_callback([e.x, e.y])

            GraphicsObject.on_triple_left_click(self)

    def _on_middle_triple_click(self, e):
        if self.is_open():
            self.mouse_triple_middle_click = e.x, e.y
            self.last_mouse_event = self.trans.world(e.x, e.y)

            if self._mouse_callback:
                self._mouse_callback([e.x, e.y])

            GraphicsObject.on_triple_middle_click(self)

    def _on_right_triple_click(self, e):
        if self.is_open():
            self.mouse_triple_right_click = e.x, e.y
            self.last_mouse_event = self.trans.world(e.x, e.y)

            if self._mouse_callback:
                self._mouse_callback([e.x, e.y])

            GraphicsObject.on_triple_right_click(self)

    # Single Click

    def _on_left_click(self, e):
        if self.is_open():
            self.mouse_left_click = e.x, e.y
            self.last_mouse_event = self.trans.world(e.x, e.y)

            self.total_left_mouse_clicks += 1

            if self._mouse_callback:
                self._mouse_callback([e.x, e.y])

            GraphicsObject.on_left_click(self)

    def _on_middle_click(self, e):
        if self.is_open():
            self.mouse_middle_click = e.x, e.y
            self.last_mouse_event = self.trans.world(e.x, e.y)

            self.total_middle_mouse_clicks += 1

            if self._mouse_callback:
                self._mouse_callback([e.x, e.y])

            GraphicsObject.on_middle_click(self)

    def _on_right_click(self, e):
        if self.is_open():
            self.mouse_right_click = e.x, e.y
            self.last_mouse_event = self.trans.world(e.x, e.y)

            self.total_right_mouse_clicks += 1

            if self._mouse_callback:
                self._mouse_callback([e.x, e.y])

            GraphicsObject.on_right_click(self)

    # Mouse Release

    def _on_left_release(self, e):
        if self.is_open():
            self.left_mouse_down = False
            self.is_dragging = False
            self._on_left_click(e)

    def _on_middle_release(self, e):
        if self.is_open():
            self.middle_mouse_down = False
            self._on_middle_click(e)

    def _on_right_release(self, e):
        if self.is_open():
            self.right_mouse_down = False
            self._on_right_click(e)

    #  Mouse Press

    def _on_left_press(self, e):
        if self.is_open():
            self.mouse_left_press = e.x, e.y
            self.last_mouse_event = self.mouse_left_press
            self.left_mouse_down = True

            if self.bounds is None or self.bounds.is_clicked([e.x, e.y]):
                self.is_dragging = True

                self.last_drag_x = e.x
                self.last_drag_y = e.y
                
            GraphicsObject.on_left_press(self)

    def _on_middle_press(self, e):
        if self.is_open():
            self.middle_mouse_down = True

            self.mouse_middle_press = e.x, e.y
            self.last_mouse_event = self.mouse_middle_press

            GraphicsObject.on_middle_press(self)

    def _on_right_press(self, e):
        if self.is_open():
            self.right_mouse_down = True

            self.mouse_right_press = e.x, e.y
            self.last_mouse_event = self.mouse_right_press

            GraphicsObject.on_right_press(self)

    # Other Mouse Functions

    def _on_mouse_motion(self, e):
        if self.is_open():
            self.mouse_pos = self.trans.world(e.x, e.y)
            if self.left_mouse_down:
                if self.is_dragging:
                    self.move(e.x - self.last_drag_x if self.is_draggable_x else 0,
                              e.y - self.last_drag_y if self.is_draggable_y else 0)

    def _on_mouse_scroll(self, e):
        self.mouse_wheel_pos = e.delta
        GraphicsObject.on_mouse_scroll(e.delta / 120, self)

    def _on_mouse_in(self, e):
        self.mouse_in_window = True

    def _on_mouse_out(self, e):
        self.mouse_in_window = False

    def start_move(self, event):
        self.x = event.x
        self.y = event.y

    def stop_move(self, event):
        self.x = None
        self.y = None

    def draw_bounds(self):
        self.bounds.draw(self)
        return self

    # WINDOW MOVING & GLIDING FUNCTIONS
    # -------------------------------------------------------------------------

    # Change position by amount
    def move(self, dx, dy):
        self.move_x(dx)
        self.move_y(dy)
        return self

    def move_x(self, dx):
        if not (isinstance(dx, int) or isinstance(dx, float)):
            raise GraphicsError("\n\nThe amount to move the window in the x-direction (dx) must be a number "
                                f"(integer or float), not {dx}")
        self.x_pos += dx
        self.master.geometry('%dx%d+%d+%d' % (self.width, self.height, self.x_pos, self.y_pos))
        return self

    def move_y(self, dy):
        if not (isinstance(dy, int) or isinstance(dy, float)):
            raise GraphicsError("\n\nThe amount to move the window in the y-direction (dy) must be a number "
                                f"(integer or float), not {dy}")
        self.y_pos += dy
        self.master.geometry('%dx%d+%d+%d' % (self.width, self.height, self.x_pos, self.y_pos))
        return self

    # Set position to
    def move_to(self, x, y):
        self.move_to_x(x)
        self.move_to_y(y)
        return self

    def move_to_x(self, x):
        if not (isinstance(x, int) or isinstance(x, float)):
            raise GraphicsError("\n\nThe x location to move the window to (x) must be a number "
                                f"(integer or float), not {x}")
        self.x_pos = x
        self.master.geometry('%dx%d+%d+%d' % (self.width, self.height, self.x_pos, self.y_pos))
        return self

    def move_to_y(self, y):
        if not (isinstance(y, int) or isinstance(y, float)):
            raise GraphicsError("\n\nThe y location to move the window to (y) must be a number "
                                f"(integer or float), not {y}")
        self.y_pos = y
        self.master.geometry('%dx%d+%d+%d' % (self.width, self.height, self.x_pos, self.y_pos))
        return self

    def move_to_point(self, p):
        if not isinstance(p, Point):
            raise GraphicsError(f"\n\nGraphicsError: point argument (p) must be a Point object, not {p}")
        self.move_to(p.x, p.y)
        return self

    # Gliding Functions

    def glide(self, dx, dy=None, time=1, easing_x=py_ease_linear(), easing_y=None):
        if dy is None:
            dy = dx
        if easing_y is None:
            easing_y = easing_x

        self.glide_x(time=time, dx=dx, easing=easing_x, _internal_call=True)
        self.glide_y(time=time, dy=dy, easing=easing_y, _internal_call=True)

        self.glide_queue.append({"Time": time, "Start": timetime(),
                                 "Update": timetime(), "Initial": [self.x_pos, self.y_pos],
                                 "Dist": [dx, dy], "EasingX": easing_x, "EasingY": easing_y})

        return self

    def glide_x(self, dx, time=1, easing=py_ease_linear(), _internal_call=False):
        if not (isinstance(dx, int) or isinstance(dx, float)):
            raise GraphicsError("\n\nThe x amount to glide the window by (dx) must be a number "
                                f"(integer or float), not {dx}")
        if not (isinstance(time, int) or isinstance(time, float)):
            raise GraphicsError("\n\nThe time to glide the window for (time) must be a number "
                                f"(integer or float), not {time}")

        if not callable(easing):
            raise GraphicsError(f"\n\nThe Easing Function Provided ({easing}) is not a valid Function")

        self.is_gliding = True
        if not _internal_call:
            start = timetime()
            initial_pos = [self.x_pos, self.y_pos]

            for glide in self.glide_queue:
                start += glide["Time"]
                initial_pos = glide["Initial"] + glide["Dist"]

            self.glide_queue.append({"Time": time, "Start": start, "Update": timetime(), "Initial": initial_pos,
                                     "Dist": [dx, 0], "EasingX": easing, "EasingY": easing})

        return self

    def glide_y(self, dy, time=1, easing=py_ease_linear(), _internal_call=False):
        if not (isinstance(dy, int) or isinstance(dy, float)):
            raise GraphicsError("\n\nThe y amount to glide the window by (dy) must be a number "
                                f"(integer or float), not {dy}")
        if not (isinstance(time, int) or isinstance(time, float)):
            raise GraphicsError("\n\nThe time to glide the window for (time) must be a number "
                                f"(integer or float), not {time}")

        if not callable(easing):
            raise GraphicsError(f"\n\nThe Easing Function Provided ({easing}) is not a valid Function")

        self.is_gliding = True
        if not _internal_call:
            start = timetime()
            initial_pos = [self.x_pos, self.y_pos]

            for glide in self.glide_queue:
                start += glide["Time"]
                initial_pos = glide["Initial"] + glide["Dist"]

            self.glide_queue.append({"Time": time, "Start": start,
                                     "Update": timetime(), "Initial": initial_pos,
                                     "Dist": [0, dy], "EasingX": easing, "EasingY": easing})

        return self

    def glide_to(self, x, y=None, time=1, easing_x=py_ease_linear(), easing_y=None):
        if y is None:
            y = x
        if easing_y is None:
            easing_y = easing_x

        self.glide_x(time=time, dx=x - self.x_pos, easing=easing_x, _internal_call=True)
        self.glide_y(time=time, dy=y - self.y_pos, easing=easing_y, _internal_call=True)

        start = timetime()
        initial_pos = [self.x_pos, self.y_pos]
        for glide in self.glide_queue:
            start += glide["Time"]
            initial_pos = glide["Initial"] + glide["Dist"]

        self.glide_queue.append({"Time": time, "Start": start, "Update": timetime(), "Initial": initial_pos,
                                 "Dist": [x - initial_pos[0], y - initial_pos[1]], "EasingX": easing_x, "EasingY": easing_y})

        return self

    def glide_to_x(self, x, time=1, easing=py_ease_linear()):
        return self.glide_x(time=time, dx=x - self.x_pos, easing=easing)

    def glide_to_y(self, y, time=1, easing=py_ease_linear()):
        return self.glide_y(time=time, dy=y - self.y_pos, easing=easing)

    def glide_to_point(self, p, time=1, easing_x=py_ease_linear(), easing_y=None):
        if not isinstance(p, Point):
            raise GraphicsError(f"\n\nGraphicsError: point argument (p) must be a Point object, not {p}")
        return self.glide_to(x=p.x, y=p.y, time=time, easing_x=easing_x, easing_y=easing_y)

    # GETTER FUNCTIONS
    # -------------------------------------------------------------------------

    # Returns the position of the window on the display of the computer
    def get_pos(self):
        return self.master.winfo_rootx(), self.master.winfo_rooty()

    def get_x_pos(self):
        return self.master.winfo_rootx()
    
    def get_y_pos(self):
        return self.master.winfo_rooty()

    def get_bk_colour(self):
        return self.bk_colour

    def get_border_width(self):
        return self.border_width

    def get_border_relief(self):
        return self.border_relief

    def get_window_cursor(self):
        return self.cursor

    def get_current_cursor(self):
        return self.current_cursor

    def is_resizable(self):
        return self.is_resizable.copy()

    def is_width_resizable(self):
        return self.is_resizable[0]

    def is_height_resizable(self):
        return self.is_resizable[1]

    def get_size(self):
        return self.get_width(), self.get_height()

    def get_width(self):
        """Return the width of the window"""
        self.width = self.master.winfo_width()
        return self.width

    def get_height(self):
        """Return the height of the window"""
        self.height = self.master.winfo_height()
        return self.height

    def get_min_size(self):
        return self.min_width, self.min_height

    def get_min_width(self):
        return self.min_width

    def get_min_height(self):
        return self.min_height

    def get_max_size(self):
        return self.max_width, self.max_height

    def get_max_width(self):
        return self.max_width

    def get_max_height(self):
        return self.max_height

    def get_title(self):
        return self.title

    def get_coords(self):
        return [self.trans.x_base, self.trans.other_y], [self.trans.other_x, self.trans.y_base]

    # SETTER FUNCTIONS
    # -------------------------------------------------------------------------

    # Sets the background colour of the window
    def set_background(self, colour):
        """Set background colour of the window"""

        if not isinstance(colour, Colour):
            if colour in STYLES[self.style].keys():
                self.bk_colour = STYLES[self.style]["background"]
            elif colour is None:
                self.bk_colour = STYLES["default"]["background"]
        else:
            self.bk_colour = colour

        self.__check_open()
        self.master.config(bg=self.bk_colour)
        return self

    def set_border_width(self, width):

        if not (isinstance(width, int) or isinstance(width, float)):
            raise GraphicsError(f"\n\nThe window's border's width must be a number (integer or float), not {width}")
        if not width > 0:
            raise GraphicsError(f"\n\nThe window's border's width must be greater than 0, not {width}")

        self.border_width = width
        self.__check_open()
        self.master.config(bd=width)
        if self.autoflush:
            self.__autoflush()
        return self

    def set_border_relief(self, relief):

        if not isinstance(relief, str):
            raise GraphicsError(f"\n\nThe border relief (relief) must be a string, not {relief}")
        if not relief.lower() in RELIEF:
            raise GraphicsError(f"\n\nThe relief for the window border must be one of {RELIEF}, not {relief}")

        self.border_relief = relief.lower()
        self.__check_open()
        self.master.config(relief=relief)
        if self.autoflush:
            self.__autoflush()
        return self

    def set_cursor(self, cursor, _internal_call=False):
        if not isinstance(cursor, str):
            raise GraphicsError(f"\n\nThe cursor must be a string, not {cursor}")
        if not cursor.lower() in CURSORS.keys():
            raise GraphicsError(f"\n\nThe cursor for the window must be one of {list(CURSORS.keys())}, not {cursor}")

        self.__check_open()
        if _internal_call:
            self.current_cursor = cursor.lower()
        else:
            self.cursor = cursor.lower()

        self.config(cursor=CURSORS[cursor])
        if self.autoflush:
            self.__autoflush()
        return self

    def set_icon(self, icon):
        if not isinstance(icon, str):
            raise GraphicsError(f"The window icon must be a string (path to .ico texture) or None, not {icon}")
        if not osisfile(f"textures/{icon}"):
            raise GraphicsError(f"The icon path you have specified ({icon}) does not exist. "
                                f"Check for spelling and make sure this is in the correct directory.")
        if not icon.endswith(".ico"):
            raise GraphicsError("The icon file must be a .ico type. "
                                "You can use an online converter to convert your file")

        self.__check_open()
        self.master.iconbitmap(icon)
        if self.autoflush:
            self.__autoflush()
        return self

    def set_title(self, title):
        if not isinstance(title, str):
            raise GraphicsError(f"\n\nThe window's title must be a string, not {title}")

        self.title = title
        self.__check_open()
        self.master.title(title)
        if self.autoflush:
            self.__autoflush()
        return self
    
    def set_draggable(self, draggable=True):
        self.is_draggable_x = draggable
        self.is_draggable_y = draggable
        return self

    def set_draggable_x(self, draggable=True):
        self.is_draggable_x = draggable
        return self

    def set_draggable_y(self, draggable=True):
        self.is_draggable_y = draggable
        return self
    
    def set_bounds(self, bounds):
        if not isinstance(bounds, GraphicsObject):
            raise GraphicsError(f"\n\nThe window's bounds must be a GraphicsObject, not {bounds}")
        self.bounds = bounds
        return self

    # Dimensions Related

    def set_width(self, width):
        if not (isinstance(width, int) or isinstance(width, float)):
            raise GraphicsError(f"\n\nThe window's width must be a number (integer or float), not {width}")
        if not width > 0:
            raise GraphicsError(f"\n\nThe window's width must be greater than 0, not {width}")

        if width < self.min_width:
            self.min_width = width
            warning = f"\n\nWindow width ({width}) is less than window's minimum width ({self.min_width}). " \
                      f"Window minimum width has been set to {width}"
            raise GraphicsError(warning)

        elif width > self.max_width:
            warning = f"\n\nWindow width ({width}) is greater than window's maximum width ({self.max_width}). " \
                      f"Window maximum width has been set to {width}."
            max_width = width
            raise GraphicsError(warning)

        self.width = width
        self.__check_open()
        self.master.config(width=self.width)
        if self.autoflush:
            self.__autoflush()
        return self

    def set_height(self, height):
        if not (isinstance(height, int) or isinstance(height, float)):
            raise GraphicsError(f"\n\nThe window's height must be a number (integer or float), not {height}")
        if not height > 0:
            raise GraphicsError(f"\n\nThe window's height must be greater than 0, not {height}")

        if height < self.min_height:
            self.min_height = height
            warning = f"\n\nWindow height (height={height}) is less than window's minimum height (min_height=" \
                      f"{self.min_height}). Window's minimum height has been set to {height}"
            raise GraphicsError(warning)

        elif height > self.max_height:
            warning = f"\n\nWindow height ({height}) is greater than window's maximum height ({self.max_height}). " \
                      f"Window maximum height has been set to {height}."
            max_height = height
            raise GraphicsError(warning)

        self.height = height
        self.__check_open()
        self.master.config(height=height)
        if self.autoflush:
            self.__autoflush()
        return self

    def set_min_height(self, min_height):
        if not (isinstance(min_height, int) or isinstance(min_height, float)):
            raise GraphicsError(f"\n\nThe window's minimum height (min_height) must be a number (integer or float), "
                                f"not {min_height}")

        if min_height > self.max_height:
            raise GraphicsError(f"Minimum height ({min_height}) is larger than the window's maximum height"
                                f"({self.max_height})")

        self.min_height = min_height
        if min_height > self.height:
            warning = "\n\nMinimum height (min_height={}) is greater than the window's current height (height={})." + \
                      "Window's height has been set to {})".format(min_height, self.height, min_height)

            raise GraphicsError(warning)

        self.master.minsize(self.min_width, self.min_height)
        return self

    def set_min_width(self, min_width):
        if not (isinstance(min_width, int) or isinstance(min_width, float)):
            raise GraphicsError(f"\n\nThe window's minimum min_width (min_width) must be a number (integer or float), "
                                f"not {min_width}")

        if min_width > self.max_width:
            raise GraphicsError(f"Minimum width ({min_width}) is larger than the window's maximum width"
                                f"({self.max_width})")

        self.min_width = min_width
        if min_width > self.width:
            warning = "\n\nMinimum min_width (min_width={}) is greater than the window's current min_width (min_width={})." + \
                      "Window's min_width has been set to {})".format(min_width, self.width, min_width)
            raise GraphicsError(warning)

        self.master.minsize(self.min_width, self.min_height)
        return self

    def set_size(self, width, height):
        self.set_height(height)
        return self.set_width(width)

    def set_min_size(self, min_width, min_height):
        self.set_min_width(min_width)
        return self.set_min_height(min_height)

    def set_max_size(self, max_width, max_height):
        self.set_max_width(max_width)
        return self.set_max_height(max_height)

    def set_max_width(self, max_width):
        if not (isinstance(max_width, int) or isinstance(max_width, float)):
            raise GraphicsError(f"\n\nThe window's maximum width (max_width) must be a number (integer or float), "
                                f"not {max_width}")

        if max_width < self.min_width:
            raise GraphicsError(f"The Maximum Width ({max_width}) is less than the Window's minimum "
                                f"width ({self.min_width})")

        if self.width > max_width:
            warning = f"\n\nWindow's width ({self.width}) is greater than maximum width ({max_width}). " \
                      f"Maximum width has been set to {self.width}."
            self.max_width = self.width
            raise GraphicsError(warning)

        self.max_width = max_width
        self.master.maxsize(self.max_width, self.max_width)
        return self

    def set_max_height(self, max_height):
        if not (isinstance(max_height, int) or isinstance(max_height, float)):
            raise GraphicsError(f"\n\nThe window's maximum height (max_height) must be a number (integer or float), "
                                f"not {max_height}")

        if max_height < self.min_height:
            raise GraphicsError(f"The Maximum Height ({max_height}) is less than the Window's minimum "
                                f"height ({self.min_height})")

        if self.height > max_height:
            warning = f"\n\nWindow's height ({self.height}) is greater than maximum height ({max_height}). " \
                      f"Maximum height has been set to {self.height}."
            self.max_height = self.height
            raise GraphicsError(warning)

        self.max_height = max_height
        self.master.maxsize(self.max_width, self.max_height)
        return self

    def set_resizable(self, resizable_width=True, resizable_height=True):
        self.set_resizable_height(resizable_height)
        return self.set_resizable_width(resizable_width)

    def set_resizable_height(self, resizable_height=True):
        if not isinstance(resizable_height, bool):
            raise GraphicsError(f"\n\nresizable_height must be a boolean, not {resizable_height}")

        self.master.resizable(self.is_resizable[0], resizable_height)
        self.is_resizable[1] = resizable_height
        return self

    def set_resizable_width(self, resizable_width=True):
        if not isinstance(resizable_width, bool):
            raise GraphicsError(f"\n\nresizable_height must be a boolean, not {resizable_width}")

        self.master.resizable(resizable_width, self.is_resizable[1])
        self.is_resizable[0] = resizable_width
        return self

    def set_coords(self, x1, y1, x2, y2):
        """Set coordinates of window to run from (x1,y1) in the
        upper-left corner to (x2,y2) in the lower-right corner."""
        if not ((isinstance(x1, int) or isinstance(x1, float)) and (isinstance(y1, int) or isinstance(y1, float)) and
                (isinstance(x2, int) or isinstance(x2, float)) and (isinstance(y2, int) or isinstance(y2, float))):
            raise GraphicsError("\n\nCoordinate Arguments must be numbers (integers or floats)")

        self.trans = Transform(self.get_width(), self.get_height(), x1, y2, x2, y1)
        self.center = [abs((x2 - x1) / 2), abs((y2 - y1) / 2)]

        return self

    # OTHER WINDOW FUNCTIONS
    # -------------------------------------------------------------------------

    def close(self):
        """Close the window"""

        if self.closed:
            return
        self.master.destroy()
        self.closed = True
        self.__autoflush()
        Window.instances.remove(self)

        if len(Window.instances) > 0:
            Window.instances[0].master.bind("<1>", lambda event: event.widget.focus_set())
            Window.instances[0].bind("<Button-1>", Window.instances[0]._on_left_click)  # These functions are called whenever these events happen
            Window.instances[0].bind("<Button-2>", Window.instances[0]._on_middle_click)
            Window.instances[0].bind("<Button-3>", Window.instances[0]._on_right_click)

            Window.instances[0].bind("<ButtonRelease-1>", Window.instances[0]._on_left_release)
            Window.instances[0].bind("<ButtonRelease-2>", Window.instances[0]._on_middle_release)
            Window.instances[0].bind("<ButtonRelease-3>", Window.instances[0]._on_right_release)

            Window.instances[0].bind("<ButtonPress-1>", Window.instances[0]._on_left_press)
            Window.instances[0].bind("<ButtonPress-2>", Window.instances[0]._on_middle_press)
            Window.instances[0].bind("<ButtonPress-3>", Window.instances[0]._on_right_press)

            Window.instances[0].bind("<Motion>", Window.instances[0]._on_mouse_motion)
            Window.instances[0].bind_all("<MouseWheel>", Window.instances[0]._on_mouse_scroll)
            Window.instances[0].bind("<Enter>", Window.instances[0]._on_mouse_in)
            Window.instances[0].bind("<Leave>", Window.instances[0]._on_mouse_out)

            Window.instances[0].bind("<Double-Button-1>", Window.instances[0]._on_left_double_click)
            Window.instances[0].bind("<Double-Button-2>", Window.instances[0]._on_middle_double_click)
            Window.instances[0].bind("<Double-Button-3>", Window.instances[0]._on_right_double_click)

            Window.instances[0].bind("<Triple-Button-1>", Window.instances[0]._on_left_triple_click)
            Window.instances[0].bind("<Triple-Button-2>", Window.instances[0]._on_middle_triple_click)
            Window.instances[0].bind("<Triple-Button-3>", Window.instances[0]._on_right_triple_click)

            Window.instances[0].bind("<KeyPress>", Window.instances[0]._on_key_press)
            Window.instances[0].bind("<Shift-KeyPress>", Window.instances[0]._on_shift_key_press)
            Window.instances[0].bind("<Alt-KeyPress>", Window.instances[0]._on_alt_key_press)
            Window.instances[0].bind("<Control-KeyPress>", Window.instances[0]._on_control_key_press)

            Window.instances[0].bind("<KeyRelease>", Window.instances[0]._on_key_release)

        return self

    def plot(self, x, y, colour=BLACK):
        """Set pixel (x,y) to the given colour"""

        if not ((isinstance(x, int) or isinstance(x, float)) and (isinstance(y, int) or isinstance(y, float))):
            raise GraphicsError("\n\nx & y position must be numbers (integers, or floats)")
        if not isinstance(colour, Colour):
            if colour in STYLES[self.style].keys():
                colour = STYLES[self.style][colour]
            else:
                raise GraphicsError(f"Colour Argument must be a Colour or string referencing a style, not {colour}")

        self.__check_open()
        xs, ys = self.to_screen(x, y)
        self.create_line(xs, ys, xs + 1, ys, fill=colour)
        self.__autoflush()
        return self

    def flush(self):
        """Update drawing to the window"""
        self.__check_open()
        self.update_idletasks()
        return self

    def update(self, rate=None, _internal_updating=False):
        if rate is None or timetime() - self.last_update > 1 / rate:
            if self.is_gliding:
                t = timetime()

                # Check if the window is gliding for every interval of time or every frame
                # To cope with lag, you might use time to glide irrespective of how many frames passed

                # Check if the window should still be gliding
                if t - self.glide_queue[0]["Start"] >= self.glide_queue[0]["Time"]:
                    self.move_to(self.glide_queue[0]["Initial"].x + self.glide_queue[0]["Dist"].x,
                                 self.glide_queue[0]["Initial"].y + self.glide_queue[0]["Dist"].y)

                    self.glide_queue.pop(0)  # Remove the object from the gliding queue
                    if len(self.glide_queue) == 0:
                        self.is_gliding = False
                else:

                    perX = self.glide_queue[0]["EasingX"]((t - self.glide_queue[0]['Start']) / self.glide_queue[0]['Time'])
                    perY = self.glide_queue[0]["EasingY"]((t - self.glide_queue[0]['Start']) / self.glide_queue[0]['Time'])

                    self.move_to(self.glide_queue[0]["Initial"].x + self.glide_queue[0]["Dist"].x * perX,
                                 self.glide_queue[0]["Initial"].y + self.glide_queue[0]["Dist"].y * perY)

                    self.glide_queue[0]["Update"] = timetime()

            GraphicsObject.on_update(self)

            tkCanvas.update(self)

            self.last_update = timetime()
        return self

    def save_canvas(self, height=None, width=None):
        x = self.master.winfo_rootx() + 20
        y = self.master.winfo_rooty() + 25

        if height is None:
            y1 = y + self.height * 5 / 4
        else:
            y1 = y + height * 5 / 4

        if width is None:
            x1 = x + self.width * 5 / 4
        else:
            x1 = x + width * 5 / 4

        self.imgs += 1
        return ImageGrab.grab().crop((x, y, x1, y1)).save("{}{}.png".format(self.title, self.imgs - 1))

    # GET MOUSE CLICK FUNCTIONS
    # -------------------------------------------------------------------------

    def get_left_mouse_press(self):
        """Wait for mouse click and return Point object representing
        the click"""

        mouse_pos = None

        while mouse_pos is None:
            mouse_pos = self.check_left_mouse_press()
            self.update()

            if self.closed:
                break

        return mouse_pos

    def get_middle_mouse_press(self):
        """Wait for mouse click and return Point object representing
        the click"""

        mouse_pos = None

        while mouse_pos is None:
            mouse_pos = self.check_middle_mouse_press()
            self.update()

            if self.closed:
                break

        return mouse_pos

    def get_right_mouse_press(self):
        """Wait for mouse click and return Point object representing
        the click"""

        mouse_pos = None

        while mouse_pos is None:
            mouse_pos = self.check_right_mouse_press()
            self.update()

            if self.closed:
                break

        return mouse_pos

    # GET MOUSE CLICK FUNCTIONS
    # -------------------------------------------------------------------------

    def get_left_mouse_click(self):
        mouse_pos = None

        while mouse_pos is None:
            mouse_pos = self.check_left_mouse_click()
            self.update()

            if self.closed:
                break

        return mouse_pos

    def get_middle_mouse_click(self):
        mouse_pos = None

        while mouse_pos is None:
            mouse_pos = self.check_middle_mouse_click()
            self.update()

            if self.closed:
                break

        return mouse_pos

    def get_right_mouse_click(self):
        mouse_pos = None

        while mouse_pos is None:
            mouse_pos = self.check_right_mouse_click()
            self.update()

            if self.closed:
                break

        return mouse_pos

    # Double Mouse Click

    def get_double_left_mouse_click(self):
        mouse_pos = None

        while mouse_pos is None:
            mouse_pos = self.check_double_left_mouse_click()
            self.update()

            if self.closed:
                break

        return mouse_pos

    def get_double_middle_mouse_click(self):
        mouse_pos = None

        while mouse_pos is None:
            mouse_pos = self.check_double_middle_mouse_click()
            self.update()

            if self.closed:
                break

        return mouse_pos

    def get_double_right_mouse_click(self):
        mouse_pos = None

        while mouse_pos is None:
            mouse_pos = self.check_double_right_mouse_click()
            self.update()

            if self.closed:
                break

        return mouse_pos

    # Triple Mouse Click

    def get_triple_left_mouse_click(self):
        mouse_pos = None

        while mouse_pos is None:
            mouse_pos = self.check_triple_left_mouse_click()
            self.update()

            if self.closed:
                break

        return mouse_pos

    def get_triple_middle_mouse_click(self):
        mouse_pos = None

        while mouse_pos is None:
            mouse_pos = self.check_triple_middle_mouse_click()
            self.update()

            if self.closed:
                break

        return mouse_pos

    def get_triple_right_mouse_click(self):
        mouse_pos = None

        while mouse_pos is None:
            mouse_pos = self.check_triple_right_mouse_click()
            self.update()

            if self.closed:
                break

        return mouse_pos

    # CHECK MOUSE FUNCTIONS
    # -------------------------------------------------------------------------

    def check_left_mouse_click(self, _refresh=True):
        if self.mouse_left_click is None or self.closed:
            return None

        x, y = self.to_world(self.mouse_left_click[0], self.mouse_left_click[1])

        if _refresh:
            self.mouse_left_click = None
        return [x, y]

    def check_middle_mouse_click(self, _refresh=True):
        if self.mouse_middle_click is None or self.closed:
            return None

        x, y = self.to_world(self.mouse_middle_click[0], self.mouse_middle_click[1])

        if _refresh:
            self.mouse_middle_click = None
        return [x, y]

    def check_right_mouse_click(self, _refresh=True):
        if self.mouse_right_click is None or self.closed:
            return None

        x, y = self.to_world(self.mouse_right_click[0], self.mouse_right_click[1])

        if _refresh:
            self.mouse_right_click = None
        return [x, y]

    # Double Mouse Clicks

    def check_double_left_mouse_click(self, _refresh=True):
        if self.mouse_double_left_click is None or self.closed:
            return None

        x, y = self.to_world(self.mouse_double_left_click[0], self.mouse_double_left_click[1])

        if _refresh:
            self.mouse_double_left_click = None
        return [x, y]

    def check_double_middle_mouse_click(self, _refresh=True):
        if self.mouse_double_middle_click is None or self.closed:
            return None

        x, y = self.to_world(self.mouse_double_middle_click[0], self.mouse_double_middle_click[1])

        if _refresh:
            self.mouse_double_middle_click = None
        return [x, y]

    def check_double_right_mouse_click(self, _refresh=True):
        if self.mouse_double_right_click is None or self.closed:
            return None

        x, y = self.to_world(self.mouse_double_right_click[0], self.mouse_double_right_click[1])

        if _refresh:
            self.mouse_double_right_click = None
        return [x, y]

    # Triple Mouse Clicks

    def check_triple_left_mouse_click(self, _refresh=True):
        if self.mouse_triple_left_click is None or self.closed:
            return None

        x, y = self.to_world(self.mouse_triple_left_click[0], self.mouse_triple_left_click[1])

        if _refresh:
            self.mouse_triple_left_click = None
        return [x, y]

    def check_triple_middle_mouse_click(self, _refresh=True):
        if self.mouse_triple_middle_click is None or self.closed:
            return None

        x, y = self.to_world(self.mouse_triple_middle_click[0], self.mouse_triple_middle_click[1])

        if _refresh:
            self.mouse_triple_middle_click = None
        return [x, y]

    def check_triple_right_mouse_click(self, _refresh=True):
        if self.mouse_triple_right_click is None or self.closed:
            return None

        x, y = self.to_world(self.mouse_triple_right_click[0], self.mouse_triple_right_click[1])

        if _refresh:
            self.mouse_triple_right_click = None
        return [x, y]

    # Mouse Press

    def check_left_mouse_press(self, _refresh=True):
        if self.mouse_left_press is None or self.closed:
            return None

        x, y = self.to_world(self.mouse_left_press[0], self.mouse_left_press[1])

        if _refresh:
            self.mouse_left_press = None
        return [x, y]

    def check_middle_mouse_press(self, _refresh=True):
        if self.mouse_left_press is None or self.closed:
            return None

        x, y = self.to_world(self.mouse_middle_press[0], self.mouse_middle_press[1])

        if _refresh:
            self.mouse_middle_press = None

        return [x, y]

    def check_right_mouse_press(self, _refresh=True):
        if self.mouse_left_press is None or self.closed:
            return None

        x, y = self.to_world(self.mouse_right_press[0], self.mouse_right_press[1])

        if _refresh:
            self.mouse_right_press = None

        return [x, y]

    # OTHER MOUSE FUNCTIONS
    # -------------------------------------------------------------------------

    def check_mouse_scroll(self):
        if self.closed:
            return None
        else:
            self.update()
            self.mouse_scroll_amount += self.mouse_wheel_pos / 120
            self.mouse_wheel_pos = 0

            return self.mouse_scroll_amount

    def get_last_mouse(self):
        if self.last_mouse_event is not None:
            return [self.last_mouse_event[0], self.last_mouse_event[1]]
        else:
            return

    def check_mouse_motion(self, _refresh=True):
        if self.mouse_pos is not None:
            x, y = self.mouse_pos

            if _refresh:
                self.mouse_pos = None

            return [x, y]
        else:
            return

    def get_mouse_position(self):
        if self.mouse_pos is not None:
            x, y = self.mouse_pos
            return [x, y]
        else:
            return

    # Mouse Down Functions

    def check_left_mouse_down(self):
        return self.left_mouse_down

    def check_right_mouse_down(self):
        return self.right_mouse_down

    def check_middle_mouse_down(self):
        return self.middle_mouse_down

    # KEYBOARD FUNCTIONS
    # -------------------------------------------------------------------------

    def get_key_press(self):
        """Wait for user to press a key and return it as a string."""
        while self.check_key_press(_refresh=False) is None:
            self.update()
        return self.check_key_press()

    def get_key_click(self):
        """Wait for user to press a key and return it as a string."""
        while not self.check_key_click(_refresh=False):
            self.update()
        return self.check_key_click()

    def get_key_with_shift_pressed(self):
        while self.check_key_with_shift_pressed(_refresh=False) is None:
            self.update()
        return self.check_key_with_shift_pressed()

    def get_key_with_alt_pressed(self):
        while self.check_key_with_alt_pressed(_refresh=False) is None:
            self.update()
        return self.check_key_with_alt_pressed()
    
    def get_key_with_control_pressed(self):
        while self.check_key_with_control_pressed(_refresh=False) is None:
            self.update()
        return self.check_key_with_control_pressed()

    def check_key_press(self, _refresh=True):
        """Return last key pressed or None if no key pressed since last call"""
        key = self.last_key_pressed
        if _refresh:
            self.last_key_pressed = None
        return key

    def check_key_click(self, _refresh=True):
        key = self.last_key_clicked
        if _refresh:
            self.last_key_clicked = None
        return key

    def check_keys_down(self):
        return self.keys_down

    def check_key_with_shift_pressed(self, _refresh=True):
        key = self.key_pressed_with_shift
        if _refresh:
            self.key_pressed_with_shift = None
        return key

    def check_key_with_alt_pressed(self, _refresh=True):
        key = self.key_pressed_with_alt
        if _refresh:
            self.key_pressed_with_alt = None
        return key

    def check_key_with_control_pressed(self, _refresh=True):
        key = self.key_pressed_with_control
        if _refresh:
            self.key_pressed_with_control = None
        return key

    def check_for_all_keys_pressed(self, *keys, _refresh=True):
        return self.keys_down.issuperset(keys)

    def check_for_all_keys_clicked(self, *keys, _refresh=True):
        return self.keys_clicked.issuperset(keys)

    def check_for_keys_pressed(self, *keys, _refresh=True):
        for key in keys:
            if key in self.keys_down:
                return True
        return False

    def check_for_keys_clicked(self, *keys, _refresh=True):
        for key in keys:
            if key in self.keys_clicked:
                return True
        return False

    # INTERNAL FUNCTIONS
    # -------------------------------------------------------------------------

    def to_screen(self, x, y):
        if not (isinstance(x, int) or isinstance(x, float)):
            raise GraphicsError(f"X coord to convert to screen coordinate must be a number (int or float), not {x}")
        if not (isinstance(y, int) or isinstance(y, float)):
            raise GraphicsError(f"Y coord to convert to screen coordinate must be a number (int or float), not {y}")
        trans = self.trans
        if trans:
            return self.trans.screen(x, y)
        else:
            return x, y

    def to_world(self, x, y):
        if not (isinstance(x, int) or isinstance(x, float)):
            raise GraphicsError(f"X coordinate to convert to world coordinate must be a number (int or float), not {x}")
        if not (isinstance(y, int) or isinstance(y, float)):
            raise GraphicsError(f"Y coordinate to convert to world coordinate must be a number (int or float), not {y}")
        trans = self.trans
        if trans:
            return self.trans.world(x, y)
        else:
            return x, y

    def add_item(self, item):
        self.items.append(item)

    def del_item(self, item):
        self.items.remove(item)

    # Redraw everything on this window
    def redraw(self):
        for item in self.items:
            item.redraw()
        if self.autoflush:
            self.update()
        return self

    # Destroying Functions

    def destroy_all(self):
        self.delete("all")
        return self

    def destroy_item(self, item):
        self.delete(item)

    def destroy_all_instances(self, obj):
        if not issubclass(obj, GraphicsObject):
            raise GraphicsError(f"\n\nGraphicsError: obj instance to destroy must be a GraphicsObject, not {obj}")

        for instance in GraphicsObject.objects.copy():
            if isinstance(instance, obj) and instance.graphwin == self:
                instance.destroy()
        return self

    def destroy_all_lines(self):
        for instance in GraphicsObject.objects.copy():
            if "Line" in instance.__repr__() and instance.graphwin == self:
                instance.destroy()
        return self

    def destroy_all_circles(self):
        for instance in GraphicsObject.objects.copy():
            if "Circle" in instance.__repr__() and instance.graphwin == self:
                instance.destroy()
        return self

    def destroy_all_rectangles(self):
        for instance in GraphicsObject.objects.copy():
            if "Rectangle" in instance.__repr__() and instance.graphwin == self:
                instance.destroy()
        return self

    def destroy_all_ovals(self):
        for instance in GraphicsObject.objects.copy():
            if "Oval" in instance.__repr__() and instance.graphwin == self:
                instance.destroy()
        return self

    def destroy_all_images(self):
        for instance in GraphicsObject.objects.copy():
            if "Image" in instance.__repr__() and instance.graphwin == self:
                instance.destroy()
        return self
    
    def destroy_all_radiobuttons(self):
        for instance in GraphicsObject.radiobutton_instances.copy():
            GraphicsObject.objects.remove(instance)
            instance.destroy()

        GraphicsObject.radiobutton_instances = {*()}
        return self
    
    def destroy_all_cyclebuttons(self):
        for instance in GraphicsObject.cyclebutton_instances.copy():
            GraphicsObject.objects.remove(instance)
            instance.destroy()
        GraphicsObject.cyclebutton_instances = {*()}
        return self
    
    def destroy_all_entries(self):
        for instance in GraphicsObject.entry_instances.copy():
            GraphicsObject.objects.remove(instance)
            instance.destroy()
        GraphicsObject.entry_instances = {*()}
        return self

    def destroy_all_buttons(self):
        for instance in GraphicsObject.button_instances.copy():
            GraphicsObject.objects.remove(instance)
            instance.destroy()
        GraphicsObject.button_instances = {*()}
        return self
    
    def destroy_all_animated_images(self):
        for instance in GraphicsObject.animated_image_instances.copy():
            GraphicsObject.objects.remove(instance)
            instance.destroy()
        GraphicsObject.animated_image_instances = {*()}
        return self


from goopylib.objects.GraphicsObject import GraphicsObject



class Transform:
    """Internal class for 2-D coordinate transformations"""

    def __init__(self, w, h, x_low, y_low, x_high, y_high):
        # w, h are width and height of window
        # (x_low,y_low) coordinates of lower-left [raw (0,h-1)]
        # (x_high,y_high) coordinates of upper-right [raw (w-1,0)]
        xspan = (x_high - x_low)
        yspan = (y_high - y_low)
        self.x_base = x_low
        self.y_base = y_high

        self.other_x = x_high
        self.other_y = y_low

        self.x_scale = xspan / w
        self.y_scale = yspan / h

    def screen(self, x, y):
        # Returns x,y in screen (actually window) coordinates
        xs = (x - self.x_base) / self.x_scale
        ys = (self.y_base - y) / self.y_scale
        return int(xs + 0.5), int(ys + 0.5)

    def world(self, xs, ys):
        # Returns xs,ys in world coordinates
        x = xs * self.x_scale + self.x_base
        y = self.y_base - ys * self.y_scale
        return x, y


class VectorEquation:
    def __init__(self, equation):
        self.equation = equation

    def __repr__(self):
        return self.equation

    def is_clicked(self, pos):
        if pos is not None:
            x = pos[0]
            y = pos[1]
            return eval(self.equation)
    
    def check_on_edge(self, pos):
        return False

    def check_overlapping(self, other):
        return False



from goopylib.math.py_bezier_curve import *

try:
    from goopylib.math.c_bezier_curve import *
except ImportError as e:
    print(f"Failed to import goopylib.math.CBezierCurve, Post this error on github.com/BhavyeMathur/goopylib/issues: {e}.")

from goopylib.math.b_spline import *

from goopylib.math.py_easing import *

from goopylib.math.triangulation import *

from goopylib.math.curves import *
from goopylib.math.interpolations import *



from goopylib.util import GraphicsError

def is_reflex(p1, p2, p3):  # check to see if 3 Points given are convex
    return not (p2[0] - p1[0]) * (p3[1] - p1[1]) - (p2[1] - p1[1]) * (p3[0] - p1[0]) >= 0


def check_in_triangle(p1, p2, p3, pt):
    if pt is None:
        return False

    # Code from https://stackoverflow.com/questions/2049582/how-to-determine-if-a-point-is-in-a-2d-triangle

    def sign(v1, v2, v3):
        return (v1[0] - v3[0]) * (v2[1] - v3[1]) - (v2[0] - v3[0]) * (v1[1] - v3[1])

    d1 = sign(pt, p1, p2)
    d2 = sign(pt, p2, p3)
    d3 = sign(pt, p3, p1)

    return not ((d1 < 0 or d2 < 0 or d3 < 0) and (d1 > 0 or d2 > 0 or d3 > 0))


def contains_no_other_points(p1, p2, p3, vertex_chain):
    for vertex in vertex_chain:
        if vertex not in (p1, p2, p3):
            if check_in_triangle(p1, p2, p3, vertex):
                return False
    return True


def is_clockwise(vertex_chain):
    # Code from https://github.com/mrbaozi/triangulation/blob/master/sources/triangulate.py
    sum = (vertex_chain[0].x - vertex_chain[len(vertex_chain) - 1].x) * \
          (vertex_chain[0].y + vertex_chain[len(vertex_chain) - 1].y)
    for i in range(len(vertex_chain) - 1):
        sum += (vertex_chain[i + 1].x - vertex_chain[i].x) * (vertex_chain[i + 1].y + vertex_chain[i].y)
    return sum > 0

# Runs in O(n^3) time
def triangulate_earclip(vertex_chain):
    if not isinstance(vertex_chain, list):
        if isinstance(vertex_chain, tuple):
            vertex_chain = list(vertex_chain)
        else:
            raise GraphicsError("\n\nGraphicsError: input argument for triangulation must be a list or tuple,"
                                f" not {vertex_chain}")
    if len(vertex_chain) == 3:
        return [tuple(vertex_chain)]
    elif len(vertex_chain) < 3:
        raise GraphicsError("\n\nGraphicsError: polygon for triangulation must have at least 3 vertices, currently, "
                            f"only {len(vertex_chain)} vertices are present")

    vertex_chain = vertex_chain.copy()

    if is_clockwise(vertex_chain):
        vertex_chain.reverse()

    triangles = []
    size = len(vertex_chain)
    while size > 3:
        for i, point in enumerate(vertex_chain):
            p1 = vertex_chain[i - 1]
            p3 = vertex_chain[(i + 1) % size]
            if not is_reflex(p1, point, p3):
                if contains_no_other_points(p1, point, p3, vertex_chain):
                    vertex_chain.remove(point)
                    size -= 1
                    triangles.append((p1, point, p3))

    triangles.append(tuple(vertex_chain))
    return triangles

# Runs in O(n^2) time
# Reference: https://www.geometrictools.com/Documentation/TriangulationByEarClipping.pdf
def triangulate_modified_earclip(vertex_chain):
    if not isinstance(vertex_chain, list):
        if isinstance(vertex_chain, tuple):
            vertex_chain = list(vertex_chain)
        else:
            raise GraphicsError("\n\nGraphicsError: input argument for triangulation must be a list or tuple,"
                                f" not {vertex_chain}")
    if len(vertex_chain) == 3:
        return [tuple(vertex_chain)]
    elif len(vertex_chain) < 3:
        raise GraphicsError("\n\nGraphicsError: polygon for triangulation must have at least 3 vertices, currently, "
                            f"only {len(vertex_chain)} vertices are present")

    vertex_chain = vertex_chain.copy()

    if is_clockwise(vertex_chain):
        vertex_chain.reverse()

    triangles = []
    ear_vertices = []

    reflex_vertices = []
    convex_vertices = []

    size = len(vertex_chain)
    # Complexity of O(n)
    i = 0
    for vertex in vertex_chain:
        if is_reflex(vertex_chain[i - 1], vertex, vertex_chain[(i + 1) % size]):
            reflex_vertices.append(vertex)
        else:
            convex_vertices.append(vertex)
        i += 1

    # Upper Limit Complexity of O(n)
    i = 0
    for vertex in convex_vertices:
        if contains_no_other_points(vertex_chain[i - 1], vertex, vertex_chain[(i + 1) % size], reflex_vertices):
            ear_vertices.append(vertex)
        i += 1

    while size > 9:
        ear_vertex = ear_vertices.pop(0)
        i = vertex_chain.index(ear_vertex)

        prev_vertex = vertex_chain[i - 1]
        next_vertex = vertex_chain[(i + 1) % size]

        vertex_chain.pop(i)
        size -= 1

        triangles.append((prev_vertex, ear_vertex, next_vertex))

        if size > 3:
            if not is_reflex(vertex_chain[i - 2], prev_vertex, next_vertex) and \
                    contains_no_other_points(vertex_chain[i - 2], prev_vertex, next_vertex, reflex_vertices):
                if prev_vertex not in ear_vertices:
                    ear_vertices.append(prev_vertex)
                    if prev_vertex in reflex_vertices:
                        reflex_vertices.remove(prev_vertex)
                        convex_vertices.append(prev_vertex)

            elif prev_vertex in ear_vertices:
                ear_vertices.remove(prev_vertex)
                if prev_vertex in convex_vertices:
                    convex_vertices.remove(prev_vertex)
                    reflex_vertices.append(prev_vertex)

            if (not is_reflex(prev_vertex, next_vertex, vertex_chain[(i + 1) % size])) and \
                    contains_no_other_points(prev_vertex, next_vertex, vertex_chain[(i + 1) % size], reflex_vertices):
                if next_vertex not in ear_vertices:
                    ear_vertices.append(next_vertex)
                    if next_vertex in reflex_vertices:
                        reflex_vertices.remove(next_vertex)
                        convex_vertices.append(next_vertex)

            elif next_vertex in ear_vertices:
                ear_vertices.remove(next_vertex)
                if next_vertex in convex_vertices:
                    convex_vertices.remove(next_vertex)
                    reflex_vertices.append(next_vertex)

        else:
            break

    triangles.append(tuple(vertex_chain))
    return triangles

# Runs in O(n log n) time
# Reference: https://cs.gmu.edu/~jmlien/teaching/09_fall_cs633/uploads/Main/lecture03.pdf
# Reference: https://www.slideserve.com/billy/brute-force-triangulation
# Work in Progress
def triangulate_plane_sweep(vertex_chain):
    if not isinstance(vertex_chain, list):
        if isinstance(vertex_chain, tuple):
            vertex_chain = list(vertex_chain)
        else:
            raise GraphicsError("\n\nGraphicsError: input argument for triangulation must be a list or tuple,"
                                f" not {vertex_chain}")
    if len(vertex_chain) == 3:
        return [tuple(vertex_chain)]
    elif len(vertex_chain) < 3:
        raise GraphicsError("\n\nGraphicsError: polygon for triangulation must have at least 3 vertices, currently, "
                            f"only {len(vertex_chain)} vertices are present")

    vertex_chain = vertex_chain.copy()
    size = len(vertex_chain)

    if is_clockwise(vertex_chain):
        vertex_chain.reverse()

    vertex = vertex_chain[-1]
    vertex2 = vertex_chain[0]

    if vertex2[1] > vertex[1]:
        last_direction = "down"
    elif vertex2[1] == vertex[1]:
        if vertex2[0] > vertex[1]:
            last_direction = "down"
        else:
            last_direction = "up"
    else:
        last_direction = "up"

    vertex_types = []

    for i, vertex in enumerate(vertex_chain):
        vertex2 = vertex_chain[(i + 1) % size]

        is_turn_vertex = False
        if vertex2[1] > vertex[1]:
            if last_direction != "down":
                is_turn_vertex = True
                last_direction = "down"

        elif vertex2[1] == vertex[1]:
            if vertex2[0] > vertex[1]:
                if last_direction != "down":
                    is_turn_vertex = True
                    last_direction = "down"
            else:
                if last_direction != "up":
                    is_turn_vertex = True
                    last_direction = "up"
        else:
            if last_direction != "up":
                is_turn_vertex = True
                last_direction = "up"

        if is_turn_vertex:
            if is_reflex(vertex_chain[i - 1], vertex, vertex2):
                if vertex2[1] < vertex[1]:
                    vertex_types.append("merge")
                elif vertex2[1] == vertex[1] and vertex2[0] > vertex[0]:
                    vertex_types.append("merge")
                else:
                    vertex_types.append("split")
            else:
                if vertex2[1] < vertex[1]:
                    vertex_types.append("end")
                elif vertex2[1] == vertex[1] and vertex2[0] > vertex[0]:
                    vertex_types.append("end")
                else:
                    vertex_types.append("start")
        else:
            vertex_types.append("regular")

    print(vertex_types)

    return triangulate_modified_earclip(vertex_chain)

# ----------------------------------------------------------------
# These are empty functions for now, but are planned in the future.

# Runs in O(n log* n) time
# Reference: http://gamma.cs.unc.edu/SEIDEL/
def triangulate_siedel(vertex_chain):
    if not isinstance(vertex_chain, list):
        if isinstance(vertex_chain, tuple):
            vertex_chain = list(vertex_chain)
        else:
            raise GraphicsError("\n\nGraphicsError: input argument for triangulation must be a list or tuple,"
                                f" not {vertex_chain}")
    if len(vertex_chain) == 3:
        return [tuple(vertex_chain)]
    elif len(vertex_chain) < 3:
        raise GraphicsError("\n\nGraphicsError: polygon for triangulation must have at least 3 vertices, currently, "
                            f"only {len(vertex_chain)} vertices are present")

# Runs in O(n log log n) time
# Reference: https://pdfs.semanticscholar.org/fe11/1fe1d1f73eb006280f4cda839f868d435f73.pdf
def triangulate_tarjan_wyk(vertex_chain):
    if not isinstance(vertex_chain, list):
        if isinstance(vertex_chain, tuple):
            vertex_chain = list(vertex_chain)
        else:
            raise GraphicsError("\n\nGraphicsError: input argument for triangulation must be a list or tuple,"
                                f" not {vertex_chain}")
    if len(vertex_chain) == 3:
        return [tuple(vertex_chain)]
    elif len(vertex_chain) < 3:
        raise GraphicsError("\n\nGraphicsError: polygon for triangulation must have at least 3 vertices, currently, "
                            f"only {len(vertex_chain)} vertices are present")

# Runs in O(n log log n) time
# Reference: https://link.springer.com/content/pdf/10.1007/BF02187846.pdf
def triangulate_wrap_around_partition(vertex_chain):
    if not isinstance(vertex_chain, list):
        if isinstance(vertex_chain, tuple):
            vertex_chain = list(vertex_chain)
        else:
            raise GraphicsError("\n\nGraphicsError: input argument for triangulation must be a list or tuple,"
                                f" not {vertex_chain}")
    if len(vertex_chain) == 3:
        return [tuple(vertex_chain)]
    elif len(vertex_chain) < 3:
        raise GraphicsError("\n\nGraphicsError: polygon for triangulation must have at least 3 vertices, currently, "
                            f"only {len(vertex_chain)} vertices are present")

# Runs in O(n) time
# Reference: https://www.cs.princeton.edu/~chazelle/pubs/polygon-triang.pdf
def triangulate_chazelle(vertex_chain):
    if not isinstance(vertex_chain, list):
        if isinstance(vertex_chain, tuple):
            vertex_chain = list(vertex_chain)
        else:
            raise GraphicsError("\n\nGraphicsError: input argument for triangulation must be a list or tuple,"
                                f" not {vertex_chain}")
    if len(vertex_chain) == 3:
        return [tuple(vertex_chain)]
    elif len(vertex_chain) < 3:
        raise GraphicsError("\n\nGraphicsError: polygon for triangulation must have at least 3 vertices, currently, "
                            f"only {len(vertex_chain)} vertices are present")



from goopylib.util import GraphicsError
import math

def distance_formula(p1, p2):
    return ((p2[0] - p1[0]) ** 2 + (p2[1] - p1[1]) ** 2) ** 0.5

def linear_curve(t, control_points):
    number_of_points = len(control_points)
    if number_of_points > 1:
        total_distance = sum(distance_formula(control_points[point - 1], control_points[point]) for point in range(number_of_points))

        points_interval = 1 / (number_of_points - 1)
        next_point = math.ceil(t / points_interval)

        mu = (t - (points_interval * (next_point - 1))) / points_interval
        mu2 = 1 - mu

        previous_point = control_points[next_point - 1]
        next_point = control_points[next_point]

        return [mu2 * previous_point[0] + mu * next_point[0], mu2 * previous_point[1] + mu * next_point[1]]
    else:
        raise GraphicsError(f"\n\nGraphicsError: must be minimum of 2 control points, not {len(control_points)}")

def cosine_curve(t, control_points):
    number_of_points = len(control_points)
    if number_of_points > 1:
        total_distance = sum(
            distance_formula(control_points[point - 1], control_points[point]) for point in range(number_of_points))

        points_interval = 1 / (number_of_points - 1)
        next_point = math.ceil(t / points_interval)

        mu = (t - (points_interval * (next_point - 1))) / points_interval

        previous_point = control_points[next_point - 1]
        next_point = control_points[next_point]

        mu2 = (1 - math.cos(mu * math.pi)) / 2

        return [(1 - mu) * previous_point[0] + mu * next_point[0], (1 - mu2) * previous_point[1] + mu2 * next_point[1]]
    else:
        raise GraphicsError(f"\n\nGraphicsError: must be minimum of 2 control points, not {len(control_points)}")
    
def cubic_curve(t, control_points):
    number_of_points = len(control_points)
    if number_of_points > 1:
        total_distance = sum(
            distance_formula(control_points[point - 1], control_points[point]) for point in range(number_of_points))

        points_interval = 1 / (number_of_points - 1)
        next_point = math.ceil(t / points_interval)

        mu = (t - (points_interval * (next_point - 1))) / points_interval

        previous_point = control_points[next_point - 1]
        previous_previous_point = control_points[next_point - 2]
        next_next_point = control_points[(next_point + 1) % number_of_points]
        next_point = control_points[next_point]

        a0 = next_next_point[1] - next_point[1] - previous_previous_point[1] + previous_point[1]
        a1 = previous_previous_point[1] - previous_point[1] - a0
        a2 = next_point[1] - previous_previous_point[1]
        a3 = previous_point[1]

        return [(1 - mu) * previous_point[0] + mu * next_point[0], (a0 * mu ** 3) + (a1 * mu ** 2) + (a2 * mu) + a3]
    else:
        raise GraphicsError(f"\n\nGraphicsError: must be minimum of 2 control points, not {len(control_points)}")

def hermite_curve(t, control_points, tension, bias):
    number_of_points = len(control_points)
    if number_of_points > 1:
        total_distance = sum(
            distance_formula(control_points[point - 1], control_points[point]) for point in range(number_of_points))

        points_interval = 1 / (number_of_points - 1)
        next_point = math.ceil(t / points_interval)

        mu = (t - (points_interval * (next_point - 1))) / points_interval

        previous_point = control_points[next_point - 1]
        previous_previous_point = control_points[next_point - 2]
        next_next_point = control_points[(next_point + 1) % number_of_points]
        next_point = control_points[next_point]

        mu2 = mu ** 2
        mu3 = mu ** 3

        m0 = (previous_point[1] - previous_previous_point[1]) * (1 + bias) * (1 - tension) / 2
        m0 += (next_point[1] - previous_point[1]) * (1 - bias) * (1 - tension) / 2
        m1 = (next_point[1] - previous_point[1]) * (1 + bias) * (1 - tension) / 2
        m1 += (next_next_point[1] - next_point[1]) * (1 - bias) * (1 - tension) / 2

        a0 = 2 * mu3 - 3 * mu2 + 1
        a1 = mu3 - 2 * mu2 + mu
        a2 = mu3 - mu2
        a3 = -2 * mu3 + 3 * mu2

        return [(1 - mu) * previous_point[0] + mu * next_point[0], a0 * previous_point[1] + a1 * m0 + a2 * m1 + a3 * next_point[1]]
    else:
        raise GraphicsError(f"\n\nGraphicsError: must be minimum of 2 control points, not {len(control_points)}")



import math

# Interpolation Easing Functions

def py_ease_linear(t=None):
    if t is None:
        def closure(time):
            return time
        return closure
    else:
        return t


def _py_ease_poly_in(t, power):
    return t ** power

def py_ease_poly_in(t=None, power=2):
    if t is None:
        def closure(time):
            return _py_ease_poly_in(time, power)
        return closure
    else:
        return t ** power


def _py_ease_poly_out(t, power):
    return 1 - (t ** power)

def py_ease_poly_out(t=None, power=2):
    if t is None:
        def closure(time):
            return _py_ease_poly_out(time, power)
        return closure
    else:
        return 1 - (t ** power)


def _py_ease_poly(t, power_in, power_out):
    t *= 2
    if t < 1:
        return t ** power_in
    else:
        return 2 - ((2 - t) ** power_out)

def py_ease_poly(t=None, power_in=2, power_out=2):
    if t is None:
        def closure(time):
            return _py_ease_poly(time, power_in, power_out)
        return closure
    else:
        t *= 2
        if t < 1:
            return t ** power_in
        else:
            return 2 - ((2 - t) ** power_out)


def _py_ease_sin_in(t):
    return 1 - math.cos(t * 1.570796327)

def py_ease_sin_in(t=None):
    if t is None:
        def closure(time):
            return _py_ease_sin_in(time)
        return closure
    else:
        return 1 - math.cos(t * 1.570796327)


def _py_ease_sin_out(t):
    return math.sin(t * 1.570796327)

def py_ease_sin_out(t=None):
    if t is None:
        def closure(time):
            return _py_ease_sin_out(time)
        return closure
    else:
        return math.sin(t * 1.570796327)


def _py_ease_sin(t):
    return 1 - math.cos(math.pi * t) / 2

def py_ease_sin(t=None):
    if t is None:
        def closure(time):
            return _py_ease_sin(time)
        return closure
    else:
        return 1 - math.cos(3.1415926535 * t) / 2


def _py_ease_circle_in(t, factor):
    return 1 - (1 - t ** 2) ** factor

def py_ease_circle_in(t=None, factor=0.5):
    if t is None:
        def closure(time):
            return _py_ease_circle_in(time, factor)
        return closure
    else:
        return 1 - (1 - t ** 2) ** factor


def _py_ease_circle_out(t, factor):
    return 1 - _py_ease_circle_in(1 - t, factor)

def py_ease_circle_out(t=None, factor=0.5):
    if t is None:
        def closure(time):
            return _py_ease_circle_out(time, factor)
        return closure
    else:
        return (1 - t ** 2) ** factor


def _py_ease_circle(t, factor_in, factor_out):
    t *= 2
    if t <= 1:
        return (1 - (1 - t * t) ** factor_in) / 2
    else:
        return ((1 - (t - 2) * (t - 2)) ** factor_out + 1) / 2

def py_ease_circle(t=None, factor_in=0.5, factor_out=0.5):
    if t is None:
        def closure(time):
            return _py_ease_circle(time, factor_in, factor_out)
        return closure
    else:
        return _py_ease_circle(t, factor_in, factor_out)


def _py_ease_back_in(t, factor):
    return t ** 2 * ((factor + 1) * t - factor)

def py_ease_back_in(t=None, factor=1.70158):
    if t is None:
        def closure(time):
            return _py_ease_back_in(time, factor)
        return closure
    else:
        return _py_ease_back_in(t, factor)


def _py_ease_back_out(t, factor):
    return 1 - (1 - t) ** 2 * ((factor + 1) * (1 - t) - factor)

def py_ease_back_out(t=None, factor=1.70158):
    if t is None:
        def closure(time):
            return _py_ease_back_out(time, factor)
        return closure
    else:
        return 1 - (1 - t) ** 2 * ((factor + 1) * (1 - t) - factor)


def _py_ease_back(t, factor_in, factor_out):
    t *= 2
    if t < 1:
        return (t * t * ((factor_in + 1) * t - factor_in)) / 2
    else:
        return ((t - 2) ** 2 * ((factor_out + 1) * (t - 2) + factor_out) + 2) / 2

def py_ease_back(t=None, factor_in=1.70158, factor_out=1.70158):
    if t is None:
        def closure(time):
            return _py_ease_back(time, factor_in, factor_out)
        return closure
    else:
        t *= 2
        if t < 1:
            return (t * t * ((factor_in + 1) * t - factor_in)) / 2
        else:
            return ((t - 2) ** 2 * ((factor_out + 1) * (t - 2) + factor_out) + 2) / 2


def _py_ease_bounce_in(t, bounces):
    return 1 - _py_ease_bounce_out(1 - t, bounces)

def py_ease_bounce_in(t=None, bounces=None):
    if t is None:
        def closure(time):
            return _py_ease_bounce_in(time, bounces)
        return closure
    else:
        return 1 - _py_ease_bounce_out(1 - t, bounces)


def _py_ease_bounce_out(t, bounces):
    if bounces is None:
        bounces = [4 / 11, 6 / 11, 8 / 11, 3 / 4, 9 / 11, 10 / 11, 15 / 16, 21 / 22, 63 / 64]
    if t < bounces[1]:
        return bounces[0] * t ** 2
    else:
        for i in range(3, len(bounces) - 2, 3):
            if t < bounces[i]:
                t -= bounces[i - 1]
                return bounces[0] * t * t + bounces[i + 1]
    t -= bounces[-2]
    return bounces[0] * t ** 2 + bounces[-1]

def py_ease_bounce_out(t=None, bounces=None):
    if bounces is None:
        bounces = [4 / 11, 6 / 11, 8 / 11, 3 / 4, 9 / 11, 10 / 11, 15 / 16, 21 / 22, 63 / 64]
    bounces.insert(0, 1 / bounces[0] / bounces[0])

    if t is None:
        def closure(time):
            return _py_ease_bounce_out(time, bounces)
        return closure
    else:
        return _py_ease_bounce_out(t, bounces)


def _py_ease_bounce(t, bounces_in, bounces_out):
    t *= 2
    if t <= 1:
        return (1 - _py_ease_bounce_in(1 - t, bounces=bounces_in)) / 2
    else:
        return (_py_ease_bounce_out(t - 1, bounces_out) + 1) / 2

def py_ease_bounce(t=None, bounces_in=None, bounces_out=None):
    if t is None:
        def closure(time):
            return _py_ease_bounce(time, bounces_in, bounces_out)
        return closure
    else:
        t *= 2
        if t <= 1:
            return (1 - _py_ease_bounce_in(1 - t, bounces=bounces_in)) / 2
        else:
            return (_py_ease_bounce_out(t - 1, bounces_out) + 1) / 2


def _py_ease_elastic_in(t, period, amplitude):
    period = period / math.tau

    a = max([amplitude, 1])
    s = math.asin(1 / a) * period

    return a * (2 ** (10 * (t - 1))) * math.sin((s - t) / period)

def py_ease_elastic_in(t=None, period=0.2, amplitude=1):
    if t is None:
        def closure(time):
            return _py_ease_elastic_in(time, period, amplitude)
        return closure
    else:
        period = period / math.tau

        a = max([amplitude, 1])
        s = math.asin(1 / a) * period

        return a * (2 ** (10 * (t - 1))) * math.sin((s - t) / period)

def _py_ease_elastic_out(t, period, amplitude):
    return 1 - _py_ease_elastic_in(1 - t, period, amplitude)

def py_ease_elastic_out(t=None, period=0.2, amplitude=1):
    if t is None:
        def closure(time):
            return _py_ease_elastic_out(time, period, amplitude)
        return closure
    else:
        return 1 - _py_ease_elastic_in(1 - t, period, amplitude)


def _py_ease_elastic(t, period_in, amplitude_in, period_out, amplitude_out):
    t = t * 2 - 1

    if t < 0:
        p = period_in / math.tau
        a = max([amplitude_in, 1])
        s = math.asin(1 / a) * p

        return a * 2 ** (10 * t) * math.sin((s - t) / p)

    else:
        p = period_out / math.tau
        a = max([amplitude_out, 1])
        s = math.asin(1 / a) * p

        return 2 - a * 2 ** (-10 * t) * math.sin((s + t) / p) / 2

def py_ease_elastic(t=None, period_in=0.2, amplitude_in=1, period_out=0.2, amplitude_out=1):
    if t is None:
        def closure(time):
            return _py_ease_elastic(time, period_in, amplitude_in, period_out, amplitude_out)
        return closure

    else:
        t = t * 2 - 1

        if t < 0:
            p = period_in / math.tau
            a = max([amplitude_in, 1])
            s = math.asin(1 / a) * p

            return a * 2 ** (10 * t) * math.sin((s - t) / p)

        else:
            p = period_out / math.tau
            a = max([amplitude_out, 1])
            s = math.asin(1 / a) * p

            return 2 - a * 2 ** (-10 * t) * math.sin((s + t) / p) / 2


def _py_ease_exponential_in(t, base):
    return base ** (10 * t - 10)

def py_ease_exponential_in(t=None, base=2):
    if t is None:
        def closure(time):
            return _py_ease_exponential_in(time, base)
        return closure
    else:
        return base ** (10 * t - 10)


def _py_ease_exponential_out(t, base):
    return 1 - (base ** (-10 * t))

def py_ease_exponential_out(t=None, base=2):
    if t is None:
        def closure(time):
            return _py_ease_exponential_out(time, base)
        return closure
    else:
        return 1 - (base ** (-10 * t))

def _py_ease_exponential(t, base_in, base_out):
    t *= 2
    if t <= 1:
        return (base_in ** (10 * t - 10)) / 2
    else:
        return (2 - base_out ** (10 - 10 * t)) / 2

def py_ease_exponential(t=None, base_in=2, base_out=2):
    if t is None:
        def closure(time):
            return _py_ease_exponential(time, base_in, base_out)
        return closure
    else:
        t *= 2
        if t <= 1:
            return (base_in ** (10 * t - 10)) / 2
        else:
            return (2 - base_out ** (10 - 10 * t)) / 2



from goopylib.Window import Window
from goopylib.objects.GraphicsObject import GraphicsObject
from goopylib.Point import Point

from goopylib.objects.Arc import Arc
from goopylib.objects.SolidArc import SolidArc
from goopylib.objects.CurvedLine import CurvedLine
from goopylib.objects.Line import Line

from goopylib.objects.Circle import Circle
from goopylib.objects.Oval import Oval
from goopylib.objects.Polygon import Polygon
from goopylib.objects.Rectangle import Rectangle

from goopylib.objects.Text import Text
from goopylib.objects.Entry import Entry
from goopylib.objects.MultilineEntry import MultilineEntry

from goopylib.objects.Image import Image
from goopylib.objects.AnimatedImage import AnimatedImage

from goopylib.objects.Checkbox import Checkbox
from goopylib.objects.Button import Button
from goopylib.objects.CycleButton import CycleButton
from goopylib.objects.RadioButton import RadioButton
from goopylib.objects.Slider import Slider
